"""
This module contains the code to generate phylogenetic networks using the
Heath model as it was called in the paper:
"Comparing the topology of phylogenetic network generators"
by Remie Janssen and Pengy Liu (2021).

This heath model is an extention of the model by Heath et al for generating trees:
"Taxon sampling affects inferences of macro-evolutionary processes from phylogenetic trees" by
Heath TA, Zwickl DJ, Kim J, Hillis DM (2008).

The main difference is that hybridization and HGT events are allowed, the rates of these events are
handled similarly to the speciation and extinction rates in the original model.
"""

import itertools
import math
import random
import sys

import networkx as nx
import numpy as np
import scipy.stats


# For each choice of reticulation arcs, calculate the distance between all pairs of taxa
# Add up all these distances for each pair, weighed by the probability of this embedded tree
# The hyb rates are e^(-hybridization_rate*sum_for_pair-offset) for each pair.
#####
# Updating distances between pairs can be done smartly:
#  for a speciation event, the distances simply increase by two times the time to speciation, and the new species copies the distances from its sister species
#  for an extinction event, the distances simply increase by two times the time to speciation, and the distances to the extinct species are removed.
#  for a HGT event, only the rates for the receiving taxon have to be updated. They can be computed from the distances to the receiving and the donating easily
#  for a hyb event, all paths go via exactly one of the parent species, so we can use those distances again.
def UpdateHybridizationRates():
    return False
    # This is done within the CalculateNetwork function


def GammaDistributionPDF(value, mean, shape):
    scale = mean / shape
    return scipy.stats.gamma.pdf(value, shape, 0, scale)


def CalculateNewRate(parent_rate, prior_mean, prior_shape, update_shape):
    proposed_rate = parent_rate * random.gammavariate(update_shape, 1 / update_shape)
    if random.random() < GammaDistributionPDF(
        proposed_rate, prior_mean, prior_shape
    ) / GammaDistributionPDF(parent_rate, prior_mean, prior_shape):
        return proposed_rate
    return parent_rate


def CalculateAllNewRates(
    parent_rates,
    update_shape,
    speciation_rate_mean,
    speciation_rate_shape,
    ext_used,
    extinction_rate_mean,
    extinction_rate_shape,
    hgt_used,
    hgt_rate_mean,
    hgt_rate_shape,
):
    sp_rate = CalculateNewRate(
        parent_rates[0], speciation_rate_mean, speciation_rate_shape, update_shape
    )
    ext_rate = 0
    if ext_used:
        ext_rate = CalculateNewRate(
            parent_rates[1], extinction_rate_mean, extinction_rate_shape, update_shape
        )
    hgt_rate = 0
    if hgt_used:
        hgt_rate = CalculateNewRate(
            parent_rates[2], hgt_rate_mean, hgt_rate_shape, update_shape
        )
    return (sp_rate, ext_rate, hgt_rate)


def DistanceToRate(
    distance,
    hybridization_left_bound,
    hybridization_right_bound,
    hybridization_left_rate,
    hybridization_right_rate,
):
    # TODO Update this rate function, make it more of a gradual switch?
    if distance <= hybridization_left_bound:
        return hybridization_left_rate
    elif distance >= hybridization_right_bound:
        return hybridization_right_rate
    return hybridization_left_rate + (
        hybridization_right_rate - hybridization_left_rate
    ) * (distance - hybridization_left_bound) / (
        hybridization_right_bound - hybridization_left_bound
    )


def RestrictToLeafSet(network, leaves_to_keep, suppress_trivial_blobs=False):
    # find leaves to remove
    leaves_to_keep = set(leaves_to_keep)
    remove_nodes = set()
    for v in network.nodes():
        if network.out_degree(v) == 0 and v not in leaves_to_keep:
            remove_nodes.add(v)

    # remove the sinks that we don't want to keep
    while remove_nodes:
        removed_node = remove_nodes.pop()
        parents = list(network.predecessors(removed_node))
        network.remove_node(removed_node)
        for p in parents:
            if network.out_degree(p) == 0:
                remove_nodes.add(p)

    # optionally: suppress biconnected components with indegree-1 and outdegree-1
    if suppress_trivial_blobs:
        # TODO: write this function.
        # What to do with the length of the resulting arc? the blob may have multiple paths, do we average?
        print("suppressing trivial blobs is not implemented yet")

    # suppress degree-2 nodes
    network = SuppressDegree2(network)
    return network


def SuppressDegree2(network):
    to_remove = []
    to_check = set(list(network.nodes())[:])
    while to_check:
        v = to_check.pop()
        if network.in_degree(v) == 1 and network.out_degree(v) == 1:
            to_remove += [v]
            parent = list(network.predecessors(v))[0]
            child = list(network.successors(v))[0]
            if network.has_edge(parent, child):
                to_check.add(parent)
                to_check.add(child)
            in_edge_length = network[parent][v]["length"]
            out_edge_length = network[v][child]["length"]
            out_edge_prob = network[v][child].get("prob")
            network.remove_edges_from([(parent, v), (v, child)])
            network.add_weighted_edges_from(
                [(parent, child, in_edge_length + out_edge_length)], weight="length"
            )
            if out_edge_prob != None:
                network[parent][child]["prob"] = out_edge_prob
    network.remove_nodes_from(to_remove)
    return network


def generate_heath_network(
    time_limit=1.0,
    taxa_limit=None,
    update_shape=2.0,
    speciation_rate_mean=2.0,
    speciation_rate_shape=2.0,
    ext_used=True,
    count_extinct=0,
    extinction_rate_mean=2.0,
    extinction_rate_shape=2.0,
    hgt_used=False,
    hgt_rate_mean=None,
    hgt_rate_shape=None,
    hgt_inheritance=0.05,
    hyb_used=False,
    hybridization_left_bound=None,
    hybridization_right_bound=None,
    hybridization_left_rate=None,
    hybridization_right_rate=None,
    simple_output=False,
):
    # Initiate the network
    nw = nx.DiGraph()
    nw.add_node(0)
    leaves = set([0])
    current_node = 1
    no_of_extinct = 0

    # Set initial rates and distances
    current_speciation_rate = random.gammavariate(
        speciation_rate_shape, speciation_rate_mean / speciation_rate_shape
    )
    current_extinction_rate = 0.0
    if ext_used:
        current_extinction_rate = random.gammavariate(
            extinction_rate_shape, extinction_rate_mean / extinction_rate_shape
        )
    current_hgt_rate = 0.0
    if hgt_used:
        current_hgt_rate = random.gammavariate(
            hgt_rate_shape, hgt_rate_mean / hgt_rate_shape
        )
    current_hybridization_rate = float(0)
    # Force the first event to be a speciation
    total_rate = current_speciation_rate
    distances = dict()

    # Set the initial leaf rates per leaf
    leaf_rates = dict()
    leaf_rates[0] = (current_speciation_rate, current_extinction_rate, current_hgt_rate)

    # Pick a time for the first event
    extra_time = np.random.exponential(1 / float(total_rate))
    current_time = extra_time

    #####First create the network as a MUL-tree, to make it easier to convert to Newick (not currently implemented)
    hybrid_nodes = dict()
    no_of_hybrids = 0

    while len(leaves) > 0 and (
        (not taxa_limit and current_time < time_limit)
        or (taxa_limit and len(leaves) + count_extinct * no_of_extinct < taxa_limit)
    ):
        random_number = random.random()
        splitting_leaf = None
        extinction_leaf = None
        hgt_donor_leaf = None
        parent_acceptor = None
        hgt_acceptor_leaf = None
        hyb_pair = None
        if random_number < current_speciation_rate / total_rate:
            ######################
            #     Speciation     #
            ######################
            if not simple_output:
                print("speciation")
            random_number = random.random() * current_speciation_rate
            for leaf, rates in leaf_rates.items():
                if random_number < rates[0]:
                    splitting_leaf = leaf
                    break
                random_number -= rates[0]
            if splitting_leaf == None:
                if not simple_output:
                    print("ouch, speciation rate computed wrong")
            nw.add_weighted_edges_from(
                [
                    (splitting_leaf, current_node, 0),
                    (splitting_leaf, current_node + 1, 0),
                ],
                weight="length",
            )
            # Update the rates and distances
            # rates
            leaf_rates[current_node] = CalculateAllNewRates(
                leaf_rates[splitting_leaf],
                update_shape,
                speciation_rate_mean,
                speciation_rate_shape,
                ext_used,
                extinction_rate_mean,
                extinction_rate_shape,
                hgt_used,
                hgt_rate_mean,
                hgt_rate_shape,
            )
            leaf_rates[current_node + 1] = CalculateAllNewRates(
                leaf_rates[splitting_leaf],
                update_shape,
                speciation_rate_mean,
                speciation_rate_shape,
                ext_used,
                extinction_rate_mean,
                extinction_rate_shape,
                hgt_used,
                hgt_rate_mean,
                hgt_rate_shape,
            )
            current_speciation_rate += (
                leaf_rates[current_node][0]
                + leaf_rates[current_node + 1][0]
                - leaf_rates[splitting_leaf][0]
            )
            current_extinction_rate += (
                leaf_rates[current_node][1]
                + leaf_rates[current_node + 1][1]
                - leaf_rates[splitting_leaf][1]
            )
            current_hgt_rate += (
                leaf_rates[current_node][2]
                + leaf_rates[current_node + 1][2]
                - leaf_rates[splitting_leaf][2]
            )
            # distances
            if hyb_used:
                for l in leaves:
                    if l != splitting_leaf:
                        pair = (splitting_leaf, l)
                        if pair in distances.keys():
                            new_distance = distances[pair]
                        else:
                            pair = (l, splitting_leaf)
                            new_distance = distances[pair]
                        distances[(l, current_node)] = new_distance
                        distances[(l, current_node + 1)] = new_distance
                        del distances[pair]
                distances[(current_node, current_node + 1)] = 0

            leaves.add(current_node)
            leaves.add(current_node + 1)
            leaves.remove(splitting_leaf)
            del leaf_rates[splitting_leaf]
            current_node += 2

        elif (
            random_number
            < (current_extinction_rate + current_speciation_rate) / total_rate
        ):
            ######################
            #     Extinction     #
            ######################
            if not simple_output:
                print("extinction")

            random_number = random.random() * current_extinction_rate
            for leaf, rates in leaf_rates.items():
                if random_number < rates[1]:
                    extinction_leaf = leaf
                    break
                random_number -= rates[1]
            if extinction_leaf == None:
                if not simple_output:
                    print("ouch, extinction rate computed wrong")

            # Update the rates and distances
            # rates
            current_speciation_rate -= leaf_rates[extinction_leaf][0]
            current_extinction_rate -= leaf_rates[extinction_leaf][1]
            current_hgt_rate -= leaf_rates[extinction_leaf][2]
            # distances
            if hyb_used:
                for l in leaves:
                    if l != extinction_leaf:
                        if (extinction_leaf, l) in distances.keys():
                            del distances[(extinction_leaf, l)]
                        else:
                            del distances[(l, extinction_leaf)]

            del leaf_rates[extinction_leaf]
            leaves.remove(extinction_leaf)
            no_of_extinct += 1

        elif (
            random_number
            < (current_extinction_rate + current_speciation_rate + current_hgt_rate)
            / total_rate
        ):
            ######################
            #      HGT event     #
            ######################
            if not simple_output:
                print("HGT")

            random_number = random.random() * current_hgt_rate
            for leaf, rates in leaf_rates.items():
                if random_number < rates[2]:
                    hgt_acceptor_leaf = leaf
                    break
                random_number -= rates[2]
            if hgt_acceptor_leaf == None:
                if not simple_output:
                    print("ouch, hgt rate computed wrong")
            if len(leaves) > 1:
                hgt_donor_leaf = random.sample(leaves - set([hgt_acceptor_leaf]), 1)[0]
                for p in nw.predecessors(hgt_acceptor_leaf):
                    parent_acceptor = p
                nw.add_weighted_edges_from(
                    [
                        (hgt_donor_leaf, hgt_acceptor_leaf, 0),
                        (hgt_donor_leaf, current_node, 0),
                        (hgt_acceptor_leaf, current_node + 1, 0),
                    ],
                    weight="length",
                )
                prob = hgt_inheritance * random.random()
                nw[parent_acceptor][hgt_acceptor_leaf]["prob"] = 1 - prob
                nw[hgt_donor_leaf][hgt_acceptor_leaf]["prob"] = prob

                hybrid_nodes[hgt_acceptor_leaf] = no_of_hybrids
                no_of_hybrids += 1
                # Update the rates and distances
                # rates
                leaf_rates[current_node + 1] = CalculateAllNewRates(
                    tuple(
                        prob * x + (1 - prob) * y
                        for x, y in zip(
                            leaf_rates[hgt_acceptor_leaf], leaf_rates[hgt_donor_leaf]
                        )
                    ),
                    update_shape,
                    speciation_rate_mean,
                    speciation_rate_shape,
                    ext_used,
                    extinction_rate_mean,
                    extinction_rate_shape,
                    hgt_used,
                    hgt_rate_mean,
                    hgt_rate_shape,
                )
                leaf_rates[current_node] = leaf_rates[hgt_donor_leaf]
                current_speciation_rate += (
                    leaf_rates[current_node + 1][0] - leaf_rates[hgt_acceptor_leaf][0]
                )
                current_extinction_rate += (
                    leaf_rates[current_node + 1][1] - leaf_rates[hgt_acceptor_leaf][1]
                )
                current_hgt_rate += (
                    leaf_rates[current_node + 1][2] - leaf_rates[hgt_acceptor_leaf][2]
                )
                del leaf_rates[hgt_donor_leaf]
                del leaf_rates[hgt_acceptor_leaf]
                # distances
                if hyb_used:
                    for l in leaves:
                        if l != hgt_acceptor_leaf:
                            acceptor_pair = (hgt_acceptor_leaf, l)
                            if acceptor_pair in distances.keys():
                                acceptor_distance = distances[acceptor_pair]
                            else:
                                acceptor_pair = (l, hgt_acceptor_leaf)
                                acceptor_distance = distances[acceptor_pair]
                            donor_pair = (l, hgt_donor_leaf)
                            donor_distance = 0.0
                            if l != hgt_donor_leaf:
                                if donor_pair in distances.keys():
                                    donor_distance = distances[donor_pair]
                                else:
                                    donor_pair = (hgt_donor_leaf, l)
                                    donor_distance = distances[donor_pair]
                                distances[(l, current_node)] = donor_distance
                                del distances[donor_pair]
                            distances[(l, current_node + 1)] = (
                                1 - prob
                            ) * acceptor_distance + prob * donor_distance
                            del distances[acceptor_pair]
                    distances[(current_node, current_node + 1)] = distances[
                        (hgt_donor_leaf, current_node + 1)
                    ]
                    del distances[(hgt_donor_leaf, current_node + 1)]
                leaves.remove(hgt_donor_leaf)
                leaves.remove(hgt_acceptor_leaf)
                leaves.add(current_node)
                leaves.add(current_node + 1)
                current_node += 2

        #        else:
        #            print("trying HGT with only one leaf")
        #             #Do nothing, there is only one leaf.

        else:
            ######################
            #    Hybridization   #
            ######################
            if not simple_output:
                print("hybridization")
            # i.e.: pick two leaf nodes, create a hybrid between these two leaves
            random_number = random.random() * current_hybridization_rate
            for pair, distance in distances.items():
                pair_rate = DistanceToRate(
                    distance,
                    hybridization_left_bound,
                    hybridization_right_bound,
                    hybridization_left_rate,
                    hybridization_right_rate,
                )
                if random_number < pair_rate:
                    hyb_pair = pair
                    break
                random_number -= pair_rate
            if hyb_pair == None and not simple_output:
                if len(leaves) == 1:
                    print("ah, no leaves for hyb")
                else:
                    print("ouch, hybridization rate computed wrong")
            nw.add_weighted_edges_from(
                [
                    (current_node, current_node + 1, 0),
                    (hyb_pair[0], current_node, 0),
                    (hyb_pair[1], current_node, 0),
                    (hyb_pair[0], current_node + 2, 0),
                    (hyb_pair[1], current_node + 3, 0),
                ],
                weight="length",
            )
            prob = random.random()
            nw[hyb_pair[0]][current_node]["prob"] = prob
            nw[hyb_pair[1]][current_node]["prob"] = 1 - prob
            hybrid_nodes[current_node] = no_of_hybrids
            no_of_hybrids += 1

            # Update the rates and distances
            # rates
            leaf_rates[current_node + 1] = CalculateAllNewRates(
                tuple(
                    prob * x + (1 - prob) * y
                    for x, y in zip(leaf_rates[hyb_pair[0]], leaf_rates[hyb_pair[1]])
                ),
                update_shape,
                speciation_rate_mean,
                speciation_rate_shape,
                ext_used,
                extinction_rate_mean,
                extinction_rate_shape,
                hgt_used,
                hgt_rate_mean,
                hgt_rate_shape,
            )
            leaf_rates[current_node + 2] = leaf_rates[hyb_pair[0]]
            leaf_rates[current_node + 3] = leaf_rates[hyb_pair[1]]
            current_speciation_rate += leaf_rates[current_node + 1][0]
            current_extinction_rate += leaf_rates[current_node + 1][1]
            current_hgt_rate += leaf_rates[current_node + 1][2]
            # distances
            # TODO FIXED?!: The order may still be wrong. It seems I already delete some distances when I still need them later.
            # These are probably the distances related to the hybrid parent species.
            for l in leaves:
                if l == hyb_pair[0]:
                    pair_0_distance = 0
                else:
                    pair_0 = (hyb_pair[0], l)
                    if pair_0 not in distances.keys():
                        pair_0 = (l, hyb_pair[0])
                    pair_0_distance = distances[pair_0]
                    distances[(l, current_node + 2)] = pair_0_distance
                #                del distances[pair_0]
                if l == hyb_pair[1]:
                    pair_1_distance = 0
                else:
                    pair_1 = (hyb_pair[1], l)
                    if pair_1 not in distances:
                        pair_1 = (l, hyb_pair[1])
                    pair_1_distance = distances[pair_1]
                    distances[(l, current_node + 3)] = pair_1_distance
                #                del distances[pair_1]
                distances[(l, current_node + 1)] = (
                    prob * pair_0_distance + (1 - prob) * pair_1_distance
                )

            if (hyb_pair[0], hyb_pair[1]) in distances:
                distances[(current_node + 2, current_node + 3)] = distances[
                    (hyb_pair[0], hyb_pair[1])
                ]
            else:
                distances[(current_node + 2, current_node + 3)] = distances[
                    (hyb_pair[1], hyb_pair[0])
                ]
            distances[(current_node + 1, current_node + 2)] = distances[
                (hyb_pair[0], current_node + 1)
            ]
            distances[(current_node + 1, current_node + 3)] = distances[
                (hyb_pair[1], current_node + 1)
            ]

            remove_pairs = []
            for pair in distances:
                if hyb_pair[0] in pair or hyb_pair[1] in pair:
                    remove_pairs += [pair]
            for pair in remove_pairs:
                del distances[pair]

            del leaf_rates[hyb_pair[0]]
            del leaf_rates[hyb_pair[1]]
            leaves.remove(hyb_pair[0])
            leaves.remove(hyb_pair[1])
            leaves.add(current_node + 1)
            leaves.add(current_node + 2)
            leaves.add(current_node + 3)
            current_node += 4

        # Now extend all pendant edges of extant taxa
        if len(leaves) == 0:
            break
        for l in leaves:
            pl = -1
            for p in nw.predecessors(l):
                pl = p
            nw[pl][l]["length"] += extra_time

        # Compute the new rates
        current_hybridization_rate = 0
        if hyb_used:
            for pair, distance in distances.items():
                distances[pair] += 2 * extra_time
                current_hybridization_rate += DistanceToRate(
                    distances[pair],
                    hybridization_left_bound,
                    hybridization_right_bound,
                    hybridization_left_rate,
                    hybridization_right_rate,
                )

        total_rate = (
            current_speciation_rate
            + current_extinction_rate
            + current_hgt_rate
            + current_hybridization_rate
        )

        # Compute the time of the next event
        extra_time = np.random.exponential(1 / total_rate)
        current_time += extra_time

    # The following corrects for overshooting the time limit
    extra_time += time_limit - current_time
    # nothing has happened yet, and there is only one node
    if len(nw) == 1:
        nw.add_weighted_edges_from([(0, 1, time_limit)], weight="length")
        leaves = set([1])
    # each leaf has a parent node, and we can extend each parent edge to time_limit
    else:
        for l in leaves:
            pl = -1
            for p in nw.predecessors(l):
                pl = p
            nw[pl][l]["length"] += extra_time

    return nw, hybrid_nodes, leaves, no_of_extinct



"""
A module for checking isomorphism between phylogenetic networks and counting automorphisms of phylogenetic networks.

Currently mostly uses the networkx isomorphism checker, extended with some specifics for phylogenetic networks 
such as labels and partial isomorphisms.
"""

from copy import deepcopy

import networkx as nx

from phylox.constants import LABEL_ATTR

#: The node attribute used to store the isometry label of a node.
ISOMETRY_LABEL_ATTR = "isometry_label"
#: The prefix used for isometry labels.
ISOMETRY_LABEL_PREFIX = "isometry_label_prefix_"
#: The prefix used for automorphism labels.
AUTOMORPHISM_LABEL_PREFIX = "automorphism_label_prefix_"

# Checks whether the nodes with the given attributes have the same label
def _same_isometry_labels(node1_attributes, node2_attributes):
    """
    Checks whether two nodes have the same label

    :param node1_attributes: the attributes of a node
    :param node2_attributes: the attributes of a node
    :return: True if the isometry label attribute ISOMETRY_LABEL_ATTR is the same, False otherwise.
    """
    return node1_attributes.get(ISOMETRY_LABEL_ATTR) == node2_attributes.get(
        ISOMETRY_LABEL_ATTR
    )


# Checks whether the nodes with the given attributes have the same label
def _same_isometry_labels_and_labels(node1_attributes, node2_attributes):
    """
    Checks whether two nodes have the same label

    :param node1_attributes: the attributes of a node
    :param node2_attributes: the attributes of a node
    :return: True if the isometry label attribute ISOMETRY_LABEL_ATTR is the same, False otherwise.
    """
    return node1_attributes.get(ISOMETRY_LABEL_ATTR) == node2_attributes.get(
        ISOMETRY_LABEL_ATTR
    ) and node1_attributes.get(LABEL_ATTR) == node2_attributes.get(LABEL_ATTR)


# Checks whether two networks are labeled isomorpgic
def is_isomorphic(network1, network2, partial_isomorphism=None, ignore_labels=False):
    """
    Determines whether two networks are labeled isomorphic.

    :param network1: a phylogenetic network, i.e., a DAG with leaf labels stored as the node attribute LABEL_ATTR.
    :param network2: a phylogenetic network, i.e., a DAG with leaf labels stored as the node attribute LABEL_ATTR.
    :return: True if the networks are labeled isomorphic, False otherwise.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.isomorphism.base import is_isomorphic
    >>> network1 = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ...     labels=[(4, "A"), (5, "B")],
    ... )
    >>> network2 = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,5),(3,6)],
    ...     labels=[(5, "B"), (6, "A")],
    ... )
    >>> is_isomorphic(network1, network2, ignore_labels=True)
    True
    >>> is_isomorphic(network1, network2, ignore_labels=False)
    False
    >>> is_isomorphic(network1, network2, partial_isomorphism=[(4,6)], ignore_labels=True)
    False
    """
    nw1 = deepcopy(network1)
    nw2 = deepcopy(network2)

    same_labels = _same_isometry_labels_and_labels
    if ignore_labels:
        same_labels = _same_isometry_labels

    partial_isomorphism = partial_isomorphism or []
    for i, corr in enumerate(partial_isomorphism):
        if not same_labels(nw1.nodes[corr[0]], nw2.nodes[corr[1]]):
            return False
        nw1.nodes[corr[0]][ISOMETRY_LABEL_ATTR] = f"{ISOMETRY_LABEL_PREFIX}{i}"
        nw2.nodes[corr[1]][ISOMETRY_LABEL_ATTR] = f"{ISOMETRY_LABEL_PREFIX}{i}"

    return nx.is_isomorphic(nw1, nw2, node_match=same_labels)


def _count_automorphisms(
    network,
    ignore_labels=False,
    partial_isomorphism=None,
    nodes_available=None,
    nodes_to_do=None,
):
    """
    Determines the number of automorphisms of a network.

    An implementation of the algorithm described in
    "A note on the graph isomorphism counting problem" by
    Rudolf Matheson in 1979.

    :param network: a phylox.DiNetwork phylogenetic network, i.e., a DAG with leaf labels.
    :param ignore_labels: if True, the automorphisms are counted without considering the labels of the nodes.
    :param partial_isomorphism: a partial isomorphism between the network and itself.
    :param nodes_available: the nodes that are available to be matched.
    :param nodes_to_do: the nodes that still need to be matched.
    :return: the number of automorphisms of the network.
    """
    nodes_available = nodes_available or []
    nodes_to_do = nodes_to_do if nodes_to_do is not None else set(network.nodes())
    same_labels = _same_isometry_labels_and_labels
    if ignore_labels:
        same_labels = _same_isometry_labels

    number_of_automorphisms = 1
    while nodes_to_do:
        node_pair_to_remove = partial_isomorphism.pop()
        node_to_remove = node_pair_to_remove[0]
        nodes_to_do.remove(node_to_remove)
        matches = 1
        for try_to_match_node in nodes_available:
            if is_isomorphic(
                network,
                network,
                partial_isomorphism=partial_isomorphism
                + [(node_to_remove, try_to_match_node)],
                ignore_labels=ignore_labels,
            ):
                matches += 1
        number_of_automorphisms *= matches
        nodes_available.append(node_to_remove)
    return number_of_automorphisms


def count_automorphisms(network, ignore_labels=False):
    """
    Determines the number of automorphisms of a network.

    :param network: a phylogenetic network, i.e., a DAG with leaf labels.
    :param ignore_labels: if True, the automorphisms are counted without considering the labels of the nodes.
    :return: the number of automorphisms of the network.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.isomorphism.base import count_automorphisms
    >>> network = DiNetwork(
    ...     edges=[(-1,0), (0, 1), (0, 2), (1, 3), (1, 4), (2, 5), (2, 6)],
    ...     labels=[(3, "A"), (4, "B"), (5, "C"), (6, "C")],
    ... )
    >>> count_automorphisms(network, ignore_labels=True)
    8
    >>> count_automorphisms(network, ignore_labels=False)
    2
    """
    partial_isomorphism = [(a, a) for a in network.nodes()]
    return _count_automorphisms(
        network,
        ignore_labels=ignore_labels,
        partial_isomorphism=partial_isomorphism,
    )



"""
A module for generating (random) LGT phylogenetic networks.

By Joan Carles Pons, Celine Scornavacca, Gabriel Cardona
With their paper: Generation of Level-k LGT Networks
PMID: 30703035 DOI: 10.1109/TCBB.2019.2895344

Adapted by RemieJanssen to output networks with a given number of leaves and reticulations
"""


import networkx as nx
import numpy as np
from networkx.utils.decorators import np_random_state, py_random_state

from phylox import DiNetwork


def last_node(net):
    return max(net.nodes())


def speciate(net, leaf):
    l = last_node(net)
    net.add_edge(leaf, l + 1)
    net.add_edge(leaf, l + 2)


def lgt(net, leaf1, leaf2):
    net.add_edge(leaf1, leaf2, secondary=True)
    l = last_node(net)
    net.add_edge(leaf1, l + 1)
    net.add_edge(leaf2, l + 2)


def leaves(net):
    return [u for u in net.nodes() if net.out_degree(u) == 0]


def internal_blobs(net):
    internal_nodes = set([u for u in net.nodes() if net.out_degree(u) > 0])
    blobs = list(nx.biconnected_components(nx.Graph(net)))
    blobs = [bl for bl in blobs if len(bl) > 2]
    nodes_in_blobs = set().union(*blobs)
    nodes_not_in_blobs = internal_nodes - nodes_in_blobs
    blobs.extend([set([u]) for u in nodes_not_in_blobs])
    return blobs


def compute_hash(net):
    mapping_blobs = {}
    blobs = internal_blobs(net)
    for blob in blobs:
        for node in blob:
            mapping_blobs[node] = blob

    mapping = {}
    for l in leaves(net):
        parent = list(net.predecessors(l))[0]
        mapping[l] = mapping_blobs[parent]
    return mapping


def internal_and_external_pairs(net):
    lvs = leaves(net)
    pairs = [(l1, l2) for l1 in lvs for l2 in lvs if l1 != l2]
    mapping = compute_hash(net)
    internal_pairs = []
    external_pairs = []
    for pair in pairs:
        if mapping[pair[0]] == mapping[pair[1]]:
            internal_pairs.append(pair)
        else:
            external_pairs.append(pair)
    return internal_pairs, external_pairs


@np_random_state("seed")
def random_pair(net, wint, wext, seed=None):
    int_pairs, ext_pairs = internal_and_external_pairs(net)
    probabilities = [wint] * len(int_pairs) + [wext] * len(ext_pairs)
    probabilities = np.array(probabilities) / sum(probabilities)
    index = seed.choice(range(len(int_pairs) + len(ext_pairs)), p=probabilities)
    return (int_pairs + ext_pairs)[index]


@np_random_state("seed")
def simulation_1(num_steps, prob_lgt, wint, wext, seed=None):
    net = nx.DiGraph()
    net.add_edge(1, 2)
    net.add_edge(1, 3)
    for i in range(num_steps):
        event = seed.choice(["spec", "lgt"], p=[1 - prob_lgt, prob_lgt])
        # event = np.random.choice(['spec','lgt'],p=[1-prob_lgt, prob_lgt])
        if event == "spec":
            l = seed.choice(leaves(net))
            speciate(net, l)
        else:
            pair = random_pair(net, wint, wext, seed=seed)
            lgt(net, pair[0], pair[1])
    return net


@np_random_state("seed")
def simulation_3(leaves_goal, retics_goal, wint, wext, seed=None):
    """
    Simulation 3 for generating networks with a given number of leaves and reticulations
    :param leaves_goal: number of leaves in the network
    :param retics_goal: number of reticulations in the network
    :param wint: weight of internal edges
    :param wext: weight of external edges
    :return: a network with the given number of leaves and reticulations
    """
    original_leaves_goal = leaves_goal
    if leaves_goal == 1:
        if retics_goal == 0:
            return DiNetwork(edges=[(0, 1)])
        # pretend we need two leaves, and connect them again later
        leaves_goal = 2
    # pick a number of extant lineages for each LGT event independently
    retics_at_lineage = dict()
    for r in range(retics_goal):
        lin = seed.choice(range(2, leaves_goal + 1))
        if lin in retics_at_lineage:
            retics_at_lineage[lin] += 1
        else:
            retics_at_lineage[lin] = 1
    network = DiNetwork(edges=[(0, 1), (1, 2), (1, 3)])
    if 2 in retics_at_lineage:
        for j in range(retics_at_lineage[2]):
            pair = random_pair(network, wint, wext, seed=seed)
            lgt(network, pair[0], pair[1])
    for i in range(3, leaves_goal + 1):
        l = seed.choice(leaves(network))
        speciate(network, l)
        if i in retics_at_lineage:
            for j in range(retics_at_lineage[i]):
                pair = random_pair(network, wint, wext, seed=seed)
                lgt(network, pair[0], pair[1])

    if original_leaves_goal == 1:
        # connect the two leaves
        unused_node = last_node(network)
        new_leaf_edge = (unused_node, unused_node + 1)
        for leaf in network.leaves:
            leaf_parent = network.parent(leaf)
            network.remove_node(leaf)
            network.add_edge(leaf_parent, unused_node)
        network._set_leaves()
    return network


def reticulations(G):
    return [v for v in G.nodes() if G.in_degree(v) == 2]


@np_random_state("seed")
def generate_network_lgt(n, k, wint=1, wext=1, max_tries=1000, seed=None):
    """
    Generate a network with a given number of leaves and reticulations
    :param n: number of leaves
    :param k: number of reticulations
    :param alpha: parameter for the weight of internal edges
    :param beta: parameter for the weight of external edges
    :param max_tries: maximum number of tries to generate a network
    :param seed: seed for the random number generator
    :return: a network with the given number of leaves and reticulations
    """

    for i in range(max_tries):
        # naive method for exact number of reticulations,
        # (original by Pons et al., not made for exact number of reticulations)
        # resG = simulation_1(n+k, alpha, 1, beta)
        # faster method:
        network = simulation_3(n, k, wint, wext, seed=seed)
        if len(reticulations(network)) == k:
            return network
    raise Exception(
        "Could not generate network with %d leaves and %d reticulations" % (n, k)
    )



import math

import networkx as nx
import numpy as np

from phylox import DiNetwork


def generate_balanced_tree(number_of_leaves):
    """
    Generates a balanced tree with the given number of leaves.

    :param number_of_leaves: the number of leaves of the balanced tree (power of 2).
    :return: a balanced tree with the given number of leaves.

    :example:
    >>> from phylox.generators.trees.well_known import generate_balanced_tree
    >>> tree = generate_balanced_tree(8)
    >>> len(tree.leaves)
    8
    """
    tree = DiNetwork()
    leaves = list(range(1, number_of_leaves + 1))
    tree.add_nodes_from(leaves)
    roots = leaves
    current_node = number_of_leaves + 1
    while len(roots) > 1:
        new_roots = []
        while len(roots) > 1:
            root1 = roots.pop()
            root2 = roots.pop()
            tree.add_edges_from([(current_node, root1), (current_node, root2)])
            new_roots.append(current_node)
            current_node += 1
        roots = new_roots
    tree.add_edge(0, roots[0])
    return tree


def generate_caterpillar(number_of_leaves):
    """
    Generates a caterpillar tree with the given number of leaves.

    :param number_of_leaves: the number of leaves of the caterpillar tree.
    :return: a caterpillar tree with the given number of leaves.

    :example:
    >>> from phylox.generators.trees.well_known import generate_caterpillar
    >>> tree = generate_caterpillar(5)
    >>> len(tree.leaves)
    5
    """
    tree = DiNetwork()
    tree.add_node(number_of_leaves + 1)
    for i in range(1, number_of_leaves):
        tree.add_edges_from(
            [
                (number_of_leaves + i, i),
                (number_of_leaves + i, number_of_leaves + i + 1),
            ]
        )
    tree = nx.relabel_nodes(tree, {2 * number_of_leaves: number_of_leaves})
    tree.add_edge(0, number_of_leaves + 1)
    return tree



"""
Functions for finding tree child sequences in a network.

Based on "Linear Time Algorithm for Tree-Child network Containment" by Remie Janssen and Yukihiro Murakami (2020)
"""

import ast
import itertools
import time
from copy import deepcopy

import networkx as nx

from phylox import DiNetwork
from phylox.cherrypicking import (
    CHERRYTYPE,
    check_reducible_pair,
    find_reducible_pairs_with_second,
    find_reticulated_cherry_with_first,
    reduce_pair,
)
from phylox.classes.dinetwork import is_tree_child
from phylox.constants import LABEL_ATTR


def find_tree_child_sequence(network, labels=False):
    """
    Find a tree child sequence for a network. If the network is orchard but not tree-child,
    this function may produce a cherry-picking sequence that reduces the network to a single edge,
    but that is not a tree-child sequence.

    :param network: The network to find a tree child sequence for.
    :type network: phylox.DiNetwork
    :param labels: Whether to return the tree child sequence as a list of labels instead of a list of nodes.
    :type labels: bool
    :return: The tree child sequence.
    :rtype: list

    :example:
    >>> from phylox import DiNetwork
    >>> N = DiNetwork(
    ...     edges=[(-1, 0), (0, 1), (0, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 5), (4, 6)],
    ...     labels=[(5, "A"), (6, "B")],
    ... )
    >>> try:
    ...     find_tree_child_sequence(N, labels=True)
    ... except ValueError:
    ...     exception_raised = True
    >>> exception_raised
    True

    >>> N = DiNetwork(
    ...     edges=[(-1, 0), (0, 1), (0, 2), (1, 2), (1, 3), (2, 4), (3, 4), (3, 5), (4, 6)],
    ...     labels=[(5, "A"), (6, "B")],
    ... )
    >>> TC_sequence = find_tree_child_sequence(N, labels=True)
    >>> TC_sequence == [("B", "A"), ("B", "A"), ("B", "A")] or TC_sequence == [("A", "B"), ("B", "A"), ("B", "A")]
    True
    """
    N = deepcopy(network)
    reducible_pairs = []
    for x in N.leaves:
        reducible_pairs.extend(find_reducible_pairs_with_second(N, x))
    tree_child_sequence = list()
    while reducible_pairs:
        pair = reducible_pairs.pop()
        cherry_type = check_reducible_pair(N, *pair)
        N, cherry_type = reduce_pair(N, *pair)
        if cherry_type != CHERRYTYPE.NONE:
            tree_child_sequence.append(pair)
            reducible_pairs.extend(find_reducible_pairs_with_second(N, pair[1]))
            reducible_pairs.extend(find_reticulated_cherry_with_first(N, pair[1]))
    if labels:
        tree_child_sequence = [
            (network.nodes[x][LABEL_ATTR], network.nodes[y][LABEL_ATTR])
            for x, y in tree_child_sequence
        ]
    if N.size() == 1:
        return tree_child_sequence
    raise ValueError("N could not be reduced to a single edge")


def check_cherry_picking_sequence(N, cherry_picking_sequence, labels=False):
    """
    Check if a cherry picking sequence is valid for a network.

    :param N: The network to check the cherry picking sequence for.
    :type N: phylox.DiNetwork
    :param cherry_picking_sequence: The cherry picking sequence to check.
    :type cherry_picking_sequence: list
    :param labels: Whether the cherry picking sequence is given as a list of labels instead of a list of nodes.
    :type labels: bool
    :return: True if the cherry picking sequence reduces the network to a single edge, False otherwise.
    :rtype: bool

    :example:
    >>> from phylox import DiNetwork
    >>> N = DiNetwork(
    ...     edges=[(-1, 0), (0, 1), (0, 2), (1, 2), (1, 3), (2, 4)],
    ...     labels=[(3, "A"), (4, "B")],
    ... )
    >>> check_cherry_picking_sequence(N, [("B", "A"), ("B", "A")], labels=True)
    True
    """
    if labels:
        cherry_picking_sequence = [
            (N.label_to_node_dict.get(x, None), N.label_to_node_dict.get(y, None))
            for x, y in cherry_picking_sequence
        ]
    for pair in cherry_picking_sequence:
        N, _ = reduce_pair(N, *pair)
    if N.size() == 1:
        return True
    return False


def tree_child_network_contains(N, M, labels=False):
    """
    Check if a tree child network N contains another network M.

    :param N: The network to check for containment.
    :type N: phylox.DiNetwork
    :param M: The network to check for.
    :type M: phylox.DiNetwork
    :param labels: Whether to return the tree child sequence as a list of labels instead of a list of nodes.
    :type labels: bool
    :return: True if N contains M, False otherwise.
    :rtype: bool

    :example:
    >>> from phylox import DiNetwork
    >>> N = DiNetwork(
    ...     edges=[(-1, 0), (0, 1), (0, 2), (1, 2), (1, 3), (2, 4)],
    ...     labels=[(3, "A"), (4, "B")],
    ... )
    >>> M = DiNetwork(
    ...     edges=[(-1, 0), (0, 1), (0, 2)],
    ...     labels=[(1, "A"), (2, "B")],
    ... )
    >>> tree_child_network_contains(N, M, labels=True)
    True
    """
    if not is_tree_child(N):
        raise ValueError("N must be a tree child network")
    return check_cherry_picking_sequence(
        M, find_tree_child_sequence(N, labels=labels), labels=labels
    )



from copy import deepcopy

from phylox.isomorphism import is_isomorphic
from phylox.rearrangement.exact_distance import ExactMethodsMixin
from phylox.rearrangement.heuristics.green_line_heuristic import HeuristicDistanceMixin
from phylox.rearrangement.move import apply_move_sequence


class RearrangementProblem(ExactMethodsMixin, HeuristicDistanceMixin):
    """
    A rearrangement problem is a tuple (N1, N2, M) where N1 and N2 are phylogenetic networks and M is a move type.

    :param network1: a phylogenetic network phylox.DiNetwork.
    :param network2: a phylogenetic network phylox.DiNetwork.
    :param move_type: a move type phylox.rearrangement.move.MoveType.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.rearrangementproblem import RearrangementProblem
    >>> from phylox.rearrangement.move import MoveType, Move
    >>> network1 = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3)],
    ...     labels=[(2, "A"), (3, "B")],
    ... )
    >>> network2 = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ...     labels=[(4, "A"), (5, "B")],
    ... )
    >>> problem = RearrangementProblem(network1, network2, MoveType.ALL)
    >>> problem.check_solution([
    ...     Move(
    ...         move_type=MoveType.VPLU,
    ...         start_edge=(1,2),
    ...         end_edge=(1,3),
    ...         network = network1,
    ...     ),
    ... ])
    True
    """

    def __init__(self, network1, network2, move_type):
        self.network1 = network1
        self.network2 = network2
        self.move_type = move_type

    def check_solution(self, seq_moves, isomorphism=None):
        """
        Checks if a sequence of moves solves the rearrangement problem.

        :param seq_moves: a sequence of moves phylox.rearrangement.move.Move.
        :param isomorphism: a partial isomorphism between the networks.
        :return: true if the sequence of moves solves the rearrangement problem, false otherwise.

        :example:
        >>> from phylox import DiNetwork
        >>> from phylox.rearrangement.rearrangementproblem import RearrangementProblem
        >>> from phylox.rearrangement.move import MoveType, Move
        >>> network1 = DiNetwork(
        ...     edges=[(0,1),(1,2),(1,3)],
        ...     labels=[(2, "A"), (3, "B")],
        ... )
        >>> network2 = DiNetwork(
        ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
        ...     labels=[(4, "A"), (5, "B")],
        ... )
        >>> problem = RearrangementProblem(network1, network2, MoveType.ALL)
        >>> problem.check_solution([
        ...     Move(
        ...         move_type=MoveType.VPLU,
        ...         start_edge=(1,2),
        ...         end_edge=(1,3),
        ...         network = network1,
        ...     ),
        ... ])
        True
        """
        if not all([move.is_type(self.move_type) for move in seq_moves]):
            return False

        final_network = apply_move_sequence(self.network1, seq_moves)

        return is_isomorphic(
            final_network, self.network2, partial_isomorphism=isomorphism
        )



# import numpy as np
# import random
# import networkx as nx
# import sys


# ##PARAMETERS
# time_limit = 0.2
# speciation_rate = 20.0
# hybridization_rate = 1.0
# inheritence = True
# edges = True
# ##


# ###############################2. I/O############################

# i = 1
# while i < len(sys.argv):
#     arg= sys.argv[i]
#     if arg == "-t":
#         i+=1
#         time_limit = float(sys.argv[i])
#     if arg == "-sp":
#         i+=1
#         speciation_rate = float(sys.argv[i])
#     if arg == "-hyb":
#         i+=1
#         hybridization_rate = float(sys.argv[i])
#     if arg == "-no_inheritence":
#         inheritence = False
#     i += 1


# nw = nx.DiGraph()
# nw.add_node(0)
# leaves = set([0])
# current_node = 1

# extra_time = np.random.exponential(1/float(speciation_rate))
# current_time = extra_time
# current_speciation_rate    = float(speciation_rate)
# current_hybridization_rate = float(0)
# rate = current_speciation_rate + current_hybridization_rate

# #First create a MUL-tree
# hybrid_nodes=dict()
# no_of_hybrids = 0

# while current_time<time_limit:
#     if random.random() < current_speciation_rate / rate:
#         #speciate
#         splitting_leaf = random.choice(list(leaves))
#         nw.add_weighted_edges_from([(splitting_leaf,current_node,0),(splitting_leaf,current_node+1,0)], weight = 'length')
#         leaves.remove(splitting_leaf)
#         leaves.add(current_node)
#         leaves.add(current_node+1)
#         current_node+=2
#     else:
#         #Hybridize
# 	#i.e.: pick two leaf nodes, merge those, and add a new leaf below this hybrid node.
#         merging = random.sample(leaves,2)
#         l0 = merging[0]
#         l1 = merging[1]
#         pl0 = -1
#         for p in nw.predecessors(l0):
#             pl0=p
#         pl1 = -1
#         for p in nw.predecessors(l1):
#             pl1=p
#         #If pl0==pl1, the new hybridization results in parallel edges.
#         if pl0 != pl1:
#             no_of_hybrids+=1
#             nw.add_weighted_edges_from([(l0,current_node,0)],weight = 'length')
#             leaves.remove(l0)
#             leaves.remove(l1)
#             leaves.add(current_node)
#             prob = random.random()
#             nw[pl0][l0]['prob'] = prob
#             nw[pl1][l1]['prob'] = 1-prob
#             hybrid_nodes[l0]=no_of_hybrids
#             hybrid_nodes[l1]=no_of_hybrids
#             current_node+=1
#     #Now extend all pendant edges
#     for l in leaves:
#         pl = -1
#         for p in nw.predecessors(l):
#             pl = p
#         nw[pl][l]['length']+=extra_time
#     no_of_leaves = len(leaves)
#     current_speciation_rate    = float(speciation_rate*no_of_leaves)
#     current_hybridization_rate = float(hybridization_rate * (no_of_leaves * (no_of_leaves - 1))/2)
#     rate = current_speciation_rate + current_hybridization_rate
#     extra_time    = np.random.exponential(1/rate)
#     current_time += extra_time


# extra_time -= current_time-time_limit
# #nothing has happened yet, and there is only one node
# if len(nw) == 1:
#     nw.add_weighted_edges_from([(0,1,time_limit)],weight = 'length')
#     leaves = set([1])
# # each leaf has a parent node, and we can extend each parent edge to time_limit
# else:
#     for l in leaves:
#         pl = -1
#         for p in nw.predecessors(l):
#             pl = p
#         nw[pl][l]['length']+=extra_time


# ############### NOW CONVERT TO NEWICK ##############

# def Newick_From_MULTree(tree,root,hybrid_nodes):
#     if tree.out_degree(root)==0:
#         if root in hybrid_nodes:
#             return "#H"+str(hybrid_nodes[root])
#         return str(root)
#     Newick = ""
#     for v in tree.successors(root):
#         Newick+= Newick_From_MULTree(tree,v,hybrid_nodes)+":"+str(tree[root][v]['length'])
#         if inheritence:
#             if v in hybrid_nodes:
#                 Newick+="::"+str(tree[root][v]['prob'])
#         Newick+= ","
#     Newick = "("+Newick[:-1]+")"
#     if root in hybrid_nodes:
#         Newick += "#H"+str(hybrid_nodes[root])
#     return Newick


# print(Newick_From_MULTree(nw,0,hybrid_nodes)+";")
# inheritence = False
# print(Newick_From_MULTree(nw,0,hybrid_nodes)+";")
# if edges:
#     for e in nw.edges(data=True):
#         info = ""
#         if e[0] in hybrid_nodes:
#             info += "#H"+str(hybrid_nodes[e[0]])
#         else:
#             info += str(e[0])
#         info += ","
#         if e[1] in hybrid_nodes:
#             info += "#H"+str(hybrid_nodes[e[1]])
#         else:
#             info += str(e[1])
#         info += ","
#         info += str(e[2]['length']) + ",1.0"
#         if 'prob' in e[2]:
#             info = info[:-3] + str(e[2]['prob'])
#         print(info)



import networkx as nx
from networkx.utils.decorators import py_random_state


# Returns all nodes below a given node (including the node itself)
def AllBelow(network, node):
    """
    Finds all nodes below a given node in a network.

    :param network: a phylogenetic network.
    :param node: a node in the network.
    :return: all nodes in the network that are below the chosen node, including this node.
    """
    lengths = nx.single_source_shortest_path_length(network, node)
    return lengths.keys()


def LowestReticAndTreeNodeAbove(network, excludedSet, allnodes=False):
    """
    Finds a list of lowest tree nodes and a list of lowest reticulation nodes above a given set of nodes.

    :param network: a phylogenetic network.
    :param excludedSet: a set of nodes of the network, must include all leaves.
    :param allnodes: a boolean value that determines whether we try to find all lowest nodes (True) or only one lowest node of each type (False, Default).
        In the latter case, we return not two lists, but two nodes.
    :return: A list of tree nodes and a list of reticulation nodes, so that each element of these lists has all their children in the excludedSet. If not allNodes, then these lists have length at most 1.
    """
    retic = None
    tree_node = None
    lowest_retics = []
    lowest_tree_nodes = []
    for node in network.nodes():
        if node not in excludedSet:
            for c in network.successors(node):
                if c not in excludedSet:
                    break
            # else runs if the loop was not ended by a break
            # this happens exactly when all of the children are in excludedSet
            else:
                if network.out_degree(node) == 2:
                    if allnodes:
                        lowest_tree_nodes += [node]
                    elif tree_node == None:
                        # For simplicity in description, take the FIRST lowest tree node that we encounter (sim. for the reticulations)
                        tree_node = node
                elif network.in_degree(node) == 2:
                    if allnodes:
                        lowest_retics += [node]
                    elif retic == None:
                        retic = node
                if not allnodes and tree_node != None and retic != None:
                    # stop if both types of lowest nodes are found
                    break
    if allnodes:
        return lowest_tree_nodes, lowest_retics
    return tree_node, retic


def HighestNodesBelow(network, excludedSet, allnodes=False):
    """
    Finds a list of highest tree nodes and a list of highest reticulation nodes below a given set of nodes.

    :param network: a phylogenetic network.
    :param excludedSet: a set of nodes of the network, must include the root.
    :param allnodes: a boolean value that determines whether we try to find all highest nodes (True) or only one highest node of each type (False, Default).
        In the latter case, we return not two lists, but two nodes.
    :return: A list of tree nodes and a list of reticulation nodes, so that each element of these lists has all their parents in the excludedSet. If not allNodes, then these lists have length at most 1.
    """
    retic = None
    tree_node = None
    highest_retics = []
    highest_tree_nodes = []
    for node in network.nodes():
        if node not in excludedSet:
            for c in network.predecessors(node):
                if c not in excludedSet:
                    break
            # else runs if the loop was not ended by a break
            # this happens exactly when all of the parents are in excludedSet
            else:
                if network.out_degree(node) == 2:
                    if allnodes:
                        highest_tree_nodes += [node]
                    elif tree_node == None:
                        # For simplicity in description, take the FIRST highest tree node that we encounter (sim. for the reticulations and leaves)
                        tree_node = node
                elif network.in_degree(node) == 2:
                    if allnodes:
                        highest_retics += [node]
                    elif retic == None:
                        retic = node
                if not allnodes and retic != None and tree_node != None:
                    # stop if all types of highest nodes are found
                    break
    if allnodes:
        return highest_tree_nodes, highest_retics
    return tree_node, retic


@py_random_state("seed")
def FindTreeNode(network, excludedSet=[], randomNodes=False, seed=None):
    """
    Finds a (random) tree node in a network.

    :param network: a phylogenetic network.
    :param excludedSet: a set of nodes of the network.
    :param randomNodes: a boolean value.
    :param seed: a seed for the random number generator.
    :return: a tree node of the network not in the excludedSet, or None if no such node exists. If randomNodes, then a tree node is selected from all candidates uniformly at random.
    """
    all_found = []
    for node in network.nodes():
        if (
            node not in excludedSet
            and network.out_degree(node) == 2
            and network.in_degree(node) == 1
        ):
            if randomNodes:
                all_found += [node]
            else:
                return node
    if all_found and randomNodes:
        return seed.choice(all_found)
    return None


@py_random_state("seed")
def FindLeaf(network, excludedSet=[], excludedParents=[], randomNodes=False, seed=None):
    """
    Finds a (random) leaf in a network.

    :param network: a phylogenetic network.
    :param excludedSet: a set of nodes of the network.
    :param excludedParents: a set of nodes of the network.
    :param randomNodes: a boolean value.
    :param seed: a seed for the random number generator.
    :return: a leaf of the network not in the excludedSet so that its parent is not in excludedParents, or None if no such node exists. If randomNodes, then a leaf is selected from all candidates uniformly at random.
    """
    all_found = []
    for node in network.nodes():
        parent = Parent(network, node)
        if (
            network.out_degree(node) == 0
            and parent not in excludedParents
            and node not in excludedSet
        ):
            if randomNodes:
                all_found += [node]
            else:
                return node
    if all_found and randomNodes:
        return seed.choice(all_found)
    return None


@py_random_state("seed")
def FindRetic(network, excludedSet=[], randomNodes=False, seed=None):
    """
    Finds a (random) reticulation in a network.

    :param network: a phylogenetic network.
    :param excludedSet: a set of nodes of the network.
    :param randomNodes: a boolean value.
    :param seed: a seed for the random number generator.
    :return: a reticulation node of the network not in the excludedSet, or None if no such node exists. If randomNodes, then a reticulation is selected from all candidates uniformly at random.
    """
    all_found = []
    for node in network.nodes():
        if node not in excludedSet and network.in_degree(node) == 2:
            if randomNodes:
                all_found += [node]
            else:
                return node
    if all_found and randomNodes:
        return seed.choice(all_found)
    return None


# # Returns a dictionary with node labels, keyed by the labels
# def Labels(network):
#     """
#     Returns the correspondence between the leaves and the leaf-labels of a given network

#     :param network: a phylogenetic network
#     :return: a dictionary, where the keys are labels and the values are nodes of the network.
#     """
#     label_dict = dict()
#     for node in network.nodes():
#         node_label = network.node[node].get('label')
#         if node_label:
#             label_dict[node_label] = node
#     return label_dict


#     # Find a sequence by choosing the move that most decreases the upper bound on the number of moves
# # This works as long as we can always decrease the bound.
# # E.g.1, this upper bound can be the length of the sequence given by Green_Line(N1,N2), the bound can always decrease after one move, if we take the move from the GL sequence (IMPLEMENTED)
# # TODO E.g.2, take the upper bound given by this algorithm with bound Green_Line
# def Deep_Dive_Scored(network1, network2, head_moves=True, bound_heuristic=Green_Line):
#     """
#     An experimental method that returns a sequence of tail/rSPR moves from network1 to network2, using the isomorphism-building heuristic for the chosen type of moves.


#     :param network1: a phylogenetic network.
#     :param network2: a phylogenetic network.
#     :param head_moves: a boolean value that determines whether head moves are used in addition to tail moves. If True we use rSPR moves, if False we use only tail moves.
#     :param bound_heuristic: a heuristic that finds a sequence between the networks quickly.
#     :return: a sequence of moves from network1 to network2.
#     """
#     if Isomorphic(network1, network2):
#         return []
#     seq = []
#     current_network = network1
#     current_best = []
#     for move in bound_heuristic(network1, network2, head_moves=head_moves):
#         current_best += [(move[0], move[1], move[3])]
#     if not current_best:
#         return False
#     done = False
#     current_length = 0
#     while not done:
#         candidate_moves = AllValidMoves(current_network, tail_moves=True, head_moves=head_moves)
#         for move in candidate_moves:
#             candidate_network = DoMove(current_network, *move)
#             if Isomorphic(candidate_network, network2):
#                 return current_best[:current_length] + [move]
#             candidate_sequence = bound_heuristic(candidate_network, network2, head_moves=head_moves)
#             if current_length + len(candidate_sequence) + 1 < len(current_best):
#                 current_best = current_best[:current_length] + [move]
#                 for move2 in candidate_sequence:
#                     current_best += [(move2[0], move2[1], move2[3])]
#         next_move = current_best[current_length]
#         current_network = DoMove(current_network, *next_move)
#         current_length += 1
#     return True



"""
This module contains some constants used throughout the package.
"""

#: The node attribute used to store the label of a node.
LABEL_ATTR = "label"
#: The edge attribute used to store the length of an edge.
LENGTH_ATTR = "length"
#: Prefix for the node name of a reticulation node when parsing a phylogenetic network from newick
RETIC_PREFIX = "__#R"



from copy import deepcopy

import networkx as nx

from phylox.cherrypicking import CHERRYTYPE, is_second_in_reducible_pair, reduce_pair
from phylox.constants import LABEL_ATTR


def is_binary(network):
    """
    Checks if the network is binary.

    :param network: a phylogenetic network phylox.DiNetwork.
    :return: true if the network is binary, false otherwise.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.classes.dinetwork import is_binary
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ... )
    >>> is_binary(network)
    True

    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(1,4)],
    ... )
    >>> is_binary(network)
    False
    """

    binary_node_types = [
        [0, 1],  # root
        [0, 2],  # root
        [1, 2],  # tree node
        [2, 1],  # reticulation
        [1, 0],  # leaf
    ]
    for node in network.nodes:
        degrees = [network.in_degree(node), network.out_degree(node)]
        if degrees not in binary_node_types:
            return False
    return True


def is_orchard(network):
    """
    Checks if the network is an orchard.

    :param network: a phylogenetic network phylox.DiNetwork.
    :return: true if the network is an orchard, false otherwise.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.classes.dinetwork import is_orchard
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ... )
    >>> is_orchard(network)
    True

    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,4),(3,5),(2,5),(3,4),(4,6),(5,7)],
    ... )
    >>> is_orchard(network)
    False
    """
    if len(network) == 0:
        return True
    leaves = network.leaves
    root = list(network.roots)[0]

    # make a copy and fix a root edge
    network_copy = deepcopy(network)
    if network_copy.out_degree(root) > 1:
        new_node = -1
        while new_node in network_copy.nodes:
            new_node -= 1
        network_copy.add_edge(new_node, root)

    # try to reduce the network copy
    done = False
    while not done:
        checked_all_leaves = True
        for leaf in leaves:
            pair = is_second_in_reducible_pair(network_copy, leaf)
            if pair:
                network_copy, cherry_type = reduce_pair(network_copy, *pair)
                if cherry_type == CHERRYTYPE.CHERRY:
                    leaves.remove(pair[0])
                checked_all_leaves = False
                break
        if len(network_copy.edges) == 1:
            return True
        done = checked_all_leaves
    return False


def is_stack_free(network):
    """
    Checks if the network is stack-free.

    :param network: a phylogenetic network phylox.DiNetwork.
    :return: true if the network is stack-free, false otherwise.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.classes.dinetwork import is_stack_free
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,4),(3,5),(2,5),(3,4),(4,6),(5,7)],
    ... )
    >>> is_stack_free(network)
    True

    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(3,5),(2,4),(4,5),(4,6),(5,7)],
    ... )
    >>> is_stack_free(network)
    False
    """
    for node in network.nodes:
        if network.is_reticulation(node) and any(
            [network.is_reticulation(child) for child in network.successors(node)]
        ):
            return False
    return True


def _is_endpoint_of_w_fence(network, node):
    if not network.is_reticulation(node):
        return False
    previous_node = node
    current_node = network.child(node)
    currently_at_fence_top = False
    while True:
        if network.is_leaf(current_node):
            return False
        if network.is_reticulation(current_node):
            if currently_at_fence_top:
                return True
            next_node = network.parent(current_node, exclude=[previous_node])
        if network.is_tree_node(current_node):
            if not currently_at_fence_top:
                return False
            next_node = network.child(current_node, exclude=[previous_node])
        previous_node, current_node = current_node, next_node
        currently_at_fence_top = not currently_at_fence_top


def is_tree_based(network):
    """
    Checks if the network is tree-based.

    :param network: a phylogenetic network phylox.DiNetwork.
    :return: true if the network is tree-based, false otherwise.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.classes.dinetwork import is_tree_based
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ... )
    >>> is_tree_based(network)
    True
    """
    if not is_binary(network):
        raise NotImplementedError(
            "tree-basedness cannot be computed for non-binary networks yet."
        )

    if len(network) > 0 and not nx.is_weakly_connected(network):
        return False

    if len(network.roots) > 1:
        return False

    for node in network.nodes:
        if _is_endpoint_of_w_fence(network, node):
            return False
    return True


def is_tree_child(network):
    """
    Checks if the network is a tree-child network.

    :param network: a phylogenetic network phylox.DiNetwork.
    :return: true if the network is a tree-child network, false otherwise.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.classes.dinetwork import is_tree_child
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ... )
    >>> is_tree_child(network)
    True

    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,4),(3,5),(2,5),(3,4),(4,6),(5,7)],
    ... )
    >>> is_tree_child(network)
    False

    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(3,5),(2,4),(4,5),(4,6),(5,7)],
    ... )
    >>> is_tree_child(network)
    False
    """

    for node in network.nodes:
        if network.is_leaf(node):
            continue
        if all([network.is_reticulation(child) for child in network.successors(node)]):
            return False
    return True


def is_leaf_labeled_single_root_network(network):
    """
    Checks if the network is a leaf-labeled network with a single root.

    :param network: a phylogenetic network phylox.DiNetwork.
    :return: a boolean value.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.classes.dinetwork import is_leaf_labeled_single_root_network
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ...     labels=[(4, "A"), (5, "B")],
    ... )
    >>> is_leaf_labeled_single_root_network(network)
    True

    >>> network = DiNetwork(
    ...     edges=[(0,1),(2,3)],
    ...     labels=[(1, "A"), (3, "B")],
    ... )
    >>> is_leaf_labeled_single_root_network(network)
    False

    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3)],
    ...     labels=[(3, "B")],
    ... )
    >>> is_leaf_labeled_single_root_network(network)
    False
    """
    rootFound = False
    for v in network.nodes:
        if network.in_degree(v) == 0:
            if rootFound:
                return False
            rootFound = True
        if network.out_degree(v) == 0 and not LABEL_ATTR in network.nodes[v]:
            return False
        if network.out_degree(v) == 1 and network.in_degree(v) == 1:
            return False
        if network.out_degree(v) > 1 and network.in_degree(v) > 1:
            return False
    return True



import random
import time
from copy import deepcopy

import networkx as nx
import numpy as np
from networkx.utils.decorators import np_random_state, py_random_state

from phylox import DiNetwork
from phylox.cherrypicking import (
    CHERRYTYPE,
    add_roots_to_sequence,
    check_reducible_pair,
    cherry_height,
    find_all_reducible_pairs,
    find_reducible_pairs_with_first,
    find_reducible_pairs_with_second,
    find_reticulated_cherry_with_first,
    get_indices_of_reducing_pairs,
    reduce_pair,
)
from phylox.constants import LABEL_ATTR, LENGTH_ATTR

# prefix for harmonized node names
HARMONIZE_NODES_BY_LABEL_PREFIX = "hnbl__"


class HybridizationProblem:
    """
    A class to represent a hybridization problem.
    I.e. a set of phylogenetic networks that need to be combined into a single phylogenetic network.

    :param list_of_networks: a list of phylogenetic networks, each given as a phylox.DiNetwork.
    :param newick_strings: if True, the input trees are given as newick strings, otherwise as phylox.DiNetworks.
    """

    def __init__(self, list_of_networks=None, newick_strings=True):
        # The dictionary of trees
        self.trees = dict()
        # the set of leaf labels of the trees
        self.leaves = set()

        # the current best sequence we have found for this set of trees
        self.best_seq = None
        # the list of reduced trees for each of the pairs in the best sequence
        self.best_red_trees = None

        # the best sequence for the algorithm using lengths as input as well
        self.best_seq_with_lengths = None
        # the sets of reduced trees for each pair in this sequence
        self.best_seq_with_lengths_red_trees = None
        # the height of each pair in this sequence
        self.best_seq_with_lengths_heights = None

        # true if distances are used
        self.distances = True

        # read the input trees in 'newick_strings'
        list_of_networks = list_of_networks or []
        for n in list_of_networks:
            if newick_strings:
                network = DiNetwork(newick=n)
            else:
                network = n
            self.trees[len(self.trees)] = network
            self.distances = self.distances and all(
                [LENGTH_ATTR in edge[2] for edge in network.edges(data=True)]
            )

        # check that the labels are unique in each tree
        # and that all leaves have a label
        # also set the leaves of the problem
        for i, tree in self.trees.items():
            rename_dict = dict()
            leaf_labels = set()
            for l in tree.leaves:
                leaf_label = tree.nodes[l][LABEL_ATTR]
                if leaf_label in leaf_labels:
                    raise ValueError(
                        "The label {} is not unique in tree {}".format(leaf_label, i)
                    )
                rename_dict[l] = HARMONIZE_NODES_BY_LABEL_PREFIX + leaf_label
                leaf_labels.add(leaf_label)
            self.leaves.update(
                [HARMONIZE_NODES_BY_LABEL_PREFIX + l for l in leaf_labels]
            )
            nx.relabel_nodes(tree, rename_dict, copy=False)
            tree._clear_cached()

    # Find new cherry-picking sequences for the trees and update the best found
    @np_random_state("seed")
    def CPSBound(
        self,
        repeats=1,
        progress=False,
        track=False,
        lengths=False,
        time_limit=None,
        seed=None,
    ):
        """
        Finds a cherry-picking sequence for the input networks, and updates the best sequence found so far.

        The method simply keeps picking (random) cherries until all networks are reduced to a single leaf.
        If lengths is True, the method picks the lowest cherry in each step (CPHeuristicLenths).
        If track is True, the method keeps track of the reducible pairs in each step, and reduces the most common pair (CPHeuristicStorePairs).
        Otherwise, the method simply picks a random cherry in each step (CPHeuristic).


        :param repeats: the number of times to run the algorithm.
        :param progress: if True, print progress information.
        :param track: if True, keep track of reducible pairs and reduce the most common pair.
        :param lengths: if True, pick the lowest cherry in each step.
        :param time_limit: if given, the algorithm stops after this many seconds.
        :return: the best sequence found so far.

        :example:
        >>> from phylox import DiNetwork
        >>> from phylox.cherrypicking.combining_networks import HybridizationProblem
        >>> network1 = DiNetwork(
        ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
        ...     labels=[(4, "A"), (5, "B")],
        ... )
        >>> network2 = DiNetwork(
        ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
        ...     labels=[(4, "A"), (5, "B")],
        ... )
        >>> problem = HybridizationProblem(
        ...     list_of_networks=[network1, network2],
        ...     newick_strings=False,
        ... )
        >>> sequence = problem.CPSBound()
        >>> sequence == [('B', 'A'), ('B', 'A')] or sequence == [('B', 'A'), ('A', 'B')]
        True
        """
        # Set the specific heuristic that we use, based on the user input and whether the trees have lengths
        Heuristic = self.CPHeuristic
        if track and not lengths:
            if progress:
                print("Tracking reducible cherries")
            Heuristic = self.CPHeuristicStorePairs
        if lengths:
            if not self.distances:
                raise ValueError("not all trees have branch lengths!")
            if progress:
                print("Picking the lowest cherry")
            Heuristic = self.CPHeuristicLengths
            heights_best = []
        # Initialize the recorded best sequences and corresponding data
        best = None
        red_trees_best = []
        starting_time = time.time()
        # Try as many times as required by the integer 'repeats'
        for i in range(repeats):
            if lengths:
                new, reduced_trees, seq_heights = Heuristic(
                    progress=progress, seed=seed
                )
                if progress:
                    print("found sequence of length: " + str(len(new)))
            else:
                new, reduced_trees = Heuristic(progress=progress, seed=seed)
                if progress:
                    print("found sequence of length: " + str(len(new)))
                    print(new)
                    print("improving sequence")
                new, reduced_trees = self.Improve_Sequence(
                    new, reduced_trees, progress=progress
                )
                if progress:
                    print("new length = " + str(len(new)))
                    print(new)
            if progress:
                print("adding roots")
            new, reduced_trees = add_roots_to_sequence(new, reduced_trees)
            if lengths:
                for i in range(len(new) - len(seq_heights)):
                    seq_heights += [seq_heights[-1]]
            if progress:
                print("final length = " + str(len(new)))
            if best == None or len(new) < len(best):
                best = new
                red_trees_best = reduced_trees
                if lengths:
                    heights_best = seq_heights
            if progress:
                print("best sequence has length " + str(len(best)))
            if time_limit and time.time() - starting_time > time_limit:
                break
        new_seq = [
            (
                x[len(HARMONIZE_NODES_BY_LABEL_PREFIX) :],
                y[len(HARMONIZE_NODES_BY_LABEL_PREFIX) :],
            )
            for (x, y) in best
        ]
        if lengths:
            if not self.best_seq_with_lengths or len(new_seq) < len(
                self.best_seq_with_lengths
            ):
                self.best_seq_with_lengths = new_seq
                self.best_seq_with_lengths_red_trees = red_trees_best
                self.best_seq_with_lengths_heights = heights_best
            return self.best_seq_with_lengths
        else:
            if not self.best_seq or len(new_seq) < len(self.best_seq):
                self.best_seq = new_seq
                self.best_red_trees = red_trees_best
            return self.best_seq

    # Version of the code that uses minimal memory: recompute reducible pairs when necessary.
    @np_random_state("seed")
    def CPHeuristic(self, progress=False, seed=None):
        if progress:
            print("Copying all inputs to reduce on")
        # Works in a copy of the input trees, copy_of_inputs, because trees have to be reduced somewhere.
        copy_of_inputs = deepcopy(self)
        if progress:
            print("Done, starting reduction of trees")
        CPS = []
        reduced_trees = []
        candidate_leaves = deepcopy(self.leaves)
        i = 1
        while copy_of_inputs.trees and i < 10:
            if progress:
                print("Sequence has length: " + str(len(CPS)))
                print(str(len(copy_of_inputs.trees)) + " trees left.\n")
                print("Reducing trivial pairs")
                # First reduce trivial cherries
            new_seq, new_red_trees = copy_of_inputs.Reduce_Trivial_Pairs(
                candidate_leaves
            )
            if progress:
                print("done")
            CPS += new_seq
            reduced_trees += new_red_trees
            if len(copy_of_inputs.trees) == 0:
                break
            # Now reduce a random cherry from a random tree
            random_index = seed.choice(list(copy_of_inputs.trees.keys()))
            random_tree = copy_of_inputs.trees[random_index]
            list_of_cherries = find_all_reducible_pairs(random_tree)

            random_cherry_index = seed.choice(range(len(list_of_cherries)))
            random_cherry = list(list_of_cherries)[random_cherry_index]
            CPS += [random_cherry]

            reduced_by_random_cherry = copy_of_inputs.Reduce_Pair_In_All(random_cherry)
            reduced_trees += [reduced_by_random_cherry]
            candidate_leaves = set(random_cherry)
            i += 1
        return CPS, reduced_trees

    # Version of the code that uses more memory: stores all reducible pairs.
    # Runs when user toggles -t or --track
    @np_random_state("seed")
    def CPHeuristicStorePairs(self, progress=False, seed=None):
        if progress:
            print("Copying all inputs to reduce on")
        # Works in a copy of the input trees, copy_of_inputs, because trees have to be reduced somewhere.
        copy_of_inputs = deepcopy(self)
        if progress:
            print("Done")
        CPS = []
        reduced_trees = []
        candidate_leaves = deepcopy(self.leaves)
        # Make dict of reducible pairs
        if progress:
            print("finding all reducible pairs")
        reducible_pairs = self.Find_All_Pairs()
        if progress:
            print("found all reducible pairs")
        while copy_of_inputs.trees:
            if progress:
                print("Sequence has length: " + str(len(CPS)))
                print(str(len(copy_of_inputs.trees)) + " trees left.\n")
                print("Reducing trivial pairs")
            # First reduce trivial cherries
            (
                new_seq,
                new_red_trees,
                reducible_pairs,
            ) = copy_of_inputs.Reduce_Trivial_Pairs_Store_Pairs(
                candidate_leaves, reducible_pairs
            )
            if progress:
                print("done")
            CPS += new_seq
            reduced_trees += new_red_trees
            if len(copy_of_inputs.trees) == 0:
                break

            # Now reduce a random cherry from a random tree
            # EITHER: (Get random tree, then random pair from the tree), just like in CPHeuristic
            random_index = seed.choice(list(copy_of_inputs.trees.keys()))
            random_tree = copy_of_inputs.trees[random_index]
            list_of_cherries = find_all_reducible_pairs(random_tree)
            random_cherry_index = seed.choice(range(len(list_of_cherries)))
            random_cherry = list(list_of_cherries)[random_cherry_index]

            # OR: (Get a random reducible pair from all pairs)
            # Note that this would result in a different algorithm than CPHeuristic, so we use the previous option
            #            random_cherry = random.choice(list(reducible_pairs.keys()))

            CPS += [random_cherry]
            # reduce all trees with this pair, this is where the list of reducible_pairs is used
            # using the list makes it faster to find all trees that need to be reduced.
            reduced_by_random_cherry = copy_of_inputs.Reduce_Pair_In_All(
                random_cherry, reducible_pairs=reducible_pairs
            )
            reducible_pairs = copy_of_inputs.Update_Reducible_Pairs(
                reducible_pairs, reduced_by_random_cherry
            )
            reduced_trees += [reduced_by_random_cherry]
            candidate_leaves = set(random_cherry)
        return CPS, reduced_trees

    # Version of the code that always picks the lowest available pair
    # Runs when user toggles -l or --lengths and all edges in the input trees have lengths.
    @np_random_state("seed")
    def CPHeuristicLengths(self, progress=False, seed=None):
        if progress:
            print("Copying all inputs to reduce on")
        # Works in a copy of the input trees, copy_of_inputs, because trees have to be reduced somewhere.
        copy_of_inputs = deepcopy(self)
        if progress:
            print("Done")
        CPS = []
        reduced_trees = []
        heights_seq = []

        candidate_leaves = deepcopy(self.leaves)
        # Make dict of reducible pairs
        if progress:
            print("finding all reducible pairs")
        reducible_pairs = self.Find_All_Pairs()
        current_heights = (
            dict()
        )  # for each reducible pair: [0] gives height, [1] the number of trees it was computed in.

        if progress:
            print("found all reducible pairs")
        while copy_of_inputs.trees:
            if progress:
                print("Sequence has length: " + str(len(CPS)))
                print(str(len(copy_of_inputs.trees)) + " trees left.\n")
                # First reduce trivial cherries
                print("Reducing trivial pairs")
            (
                new_seq,
                new_red_trees,
                reducible_pairs,
                new_heights_seq,
            ) = copy_of_inputs.Reduce_Trivial_Pairs_Lengths(
                candidate_leaves, reducible_pairs
            )
            if progress:
                print("done")
            CPS += new_seq
            reduced_trees += new_red_trees
            heights_seq += new_heights_seq
            if len(copy_of_inputs.trees) == 0:
                break

            # Now find the lowest cherry.
            current_heights = copy_of_inputs.Update_Heights(
                current_heights, reducible_pairs
            )
            lowest_cherry = None
            lowest_height = None
            lowest_height_tuple = None
            lowest_heights_found = 1
            for pair in reducible_pairs:
                height_pair_tuple = current_heights[pair][0]
                height_pair = float(height_pair_tuple[0] + height_pair_tuple[1]) / 2
                new_found = False
                if (not lowest_height) or lowest_height > height_pair:
                    new_found = True
                    lowest_heights_found = 1
                elif lowest_height == height_pair:
                    lowest_heights_found += 1
                    if seed.random() < 1 / float(lowest_heights_found):
                        new_found = True
                if new_found:
                    lowest_cherry = pair
                    lowest_height = height_pair
                    lowest_height_tuple = height_pair_tuple

            CPS += [lowest_cherry]
            heights_seq += [lowest_height_tuple]
            reduced_by_lowest_cherry = copy_of_inputs.Reduce_Pair_In_All(
                lowest_cherry, reducible_pairs=reducible_pairs
            )
            reducible_pairs = copy_of_inputs.Update_Reducible_Pairs(
                reducible_pairs, reduced_by_lowest_cherry
            )
            reduced_trees += [reduced_by_lowest_cherry]
            candidate_leaves = set(lowest_cherry)
        return CPS, reduced_trees, heights_seq

    def Update_Heights(self, current_heights, reducible_pairs):
        """
        Returns an updated dictionary of heights of the reducible pairs
        """
        for pair, trees in reducible_pairs.items():
            # updating is only necessary when the set of trees for that pair is changed or the reducible pair was not reducible before.
            if not pair in current_heights or not current_heights[pair][1] == len(
                trees
            ):
                height_pair = self.Height_Pair(pair, trees)
                current_heights[pair] = (height_pair, len(trees))
        return current_heights

    def Height_Pair(self, pair, trees):
        """
        Returns the average height of a pair in a set of trees
        The pair must be reducible in each tree in 'trees'
        """
        height_pair = [0, 0]
        for t in trees:
            height_in_t = cherry_height(self.trees[t], *pair)
            height_pair[0] += height_in_t[0]
            height_pair[1] += height_in_t[1]
        return [height_pair[0] / float(len(trees)), height_pair[1] / float(len(trees))]

    def Find_All_Pairs(self):
        """
        Finds the set of reducible pairs in all trees
        Returns a dictionary with reducible pairs as keys, and the trees they reduce as values.
        """
        reducible_pairs = dict()
        for i, t in self.trees.items():
            red_pairs_t = find_all_reducible_pairs(t)
            for pair in red_pairs_t:
                if pair not in reducible_pairs:
                    reducible_pairs[pair] = set()
                reducible_pairs[pair].add(i)
        return reducible_pairs

    def Update_Reducible_Pairs(self, reducible_pairs, new_red_trees):
        """
        Returns the updated dictionary of reducible pairs in all trees after a reduction (with the trees they reduce as values)
        we only need to update for the trees that got reduced: 'new_red_treed'
        """
        # Remove trees to update from all pairs
        for pair, trees in list(reducible_pairs.items()):
            trees.difference_update(new_red_trees)
            if len(trees) == 0:
                del reducible_pairs[pair]
        # Add the trees to the right pairs again
        for index in new_red_trees:
            if index in self.trees:
                t = self.trees[index]
                red_pairs_t = find_all_reducible_pairs(t)
                for pair in red_pairs_t:
                    if pair in reducible_pairs:
                        reducible_pairs[pair].add(index)
                    else:
                        reducible_pairs[pair] = set([index])
        return reducible_pairs

    def Reduce_Pair_In_All(self, pair, reducible_pairs=dict()):
        """
        Reduces the given pair in all networks in the problem.
        Returns the set of networks that were reduced.

        :note: This method changes the set of networks in the problem, so it should only be called in a copy of the problem.

        :param pair: a pair of leaves.
        :param reducible_pairs: a dictionary of reducible pairs and the networks in which they are reducible, as returned by Find_All_Pairs.
        """

        reduced_trees_for_pair = []
        if pair in reducible_pairs:
            trees_to_reduce = reducible_pairs[pair]
        else:
            # pair not found, trying all trees
            trees_to_reduce = deepcopy(self.trees)
        for i in trees_to_reduce:
            if i in self.trees:
                t = self.trees[i]
                t, cherry_type = reduce_pair(t, *pair, inplace=True)
                if cherry_type == CHERRYTYPE.RETICULATEDCHERRY:
                    reduced_trees_for_pair += [i]
                elif cherry_type == CHERRYTYPE.CHERRY:
                    reduced_trees_for_pair += [i]
                    t.leaves.remove(pair[0])
                if len(t.edges()) <= 1:
                    del self.trees[i]
        return set(reduced_trees_for_pair)

    def Reduce_Trivial_Pairs(self, candidate_leaves):
        """
        Reduces the trivial pairs in the current set of networks.
        Runs efficiently by giving a set of leaves 'candidate_leaves' that may be involved in trivial pairs.
        This set must be given; after a reduction of the pair (a,b) only using the leaves a and b works.
        Returns the reduced pairs and the sets of networks that were reduced.

        :note: This method changes the set of networks in the problem, so it should only be called in a copy of the problem.
        """
        seq = []
        reduced_tree_sets = []
        while candidate_leaves:
            l = candidate_leaves.pop()
            new_pairs = list(self.Trivial_Pair_With(l))
            if new_pairs:
                seq += new_pairs
                for p in new_pairs:
                    red_trees_p = self.Reduce_Pair_In_All(p)
                    reduced_tree_sets += [red_trees_p]
                    candidate_leaves = candidate_leaves | set(p)
        return seq, reduced_tree_sets

    def Reduce_Trivial_Pairs_Store_Pairs(self, candidate_leaves, reducible_pairs):
        """
        Reduces the trivial pairs in the current set of networks.
        Runs efficiently by giving a set of leaves 'candidate_leaves' that may be involved in trivial pairs.
        This set must be given; after a reduction of the pair (a,b) only using the leaves a and b works.
        Returns the reduced pairs and the sets of networks that were reduced, also updates the reducible pairs.

        This method is similar to Reduce_Trivial_Pairs, but it uses and updates the dictionary of reducible pairs to reduce the trivial pairs.

        :note: This method changes the set of networks in the problem, so it should only be called in a copy of the problem.
        """
        seq = []
        reduced_tree_sets = []
        while candidate_leaves:
            l = candidate_leaves.pop()
            new_pairs = list(self.Trivial_Pair_With(l))
            if new_pairs:
                seq += new_pairs
                for p in new_pairs:
                    red_trees_p = self.Reduce_Pair_In_All(
                        p, reducible_pairs=reducible_pairs
                    )
                    reducible_pairs = self.Update_Reducible_Pairs(
                        reducible_pairs, red_trees_p
                    )
                    reduced_tree_sets += [red_trees_p]
                    candidate_leaves = candidate_leaves | set(p)
        return seq, reduced_tree_sets, reducible_pairs

    def Reduce_Trivial_Pairs_Lengths(self, candidate_leaves, reducible_pairs):
        """
        Reduces the trivial pairs in the current set of networks.
        Runs efficiently by giving a set of leaves 'candidate_leaves' that may be involved in trivial pairs.
        This set must be given; after a reduction of the pair (a,b) only using the leaves a and b works.
        Returns the reduced pairs and the sets of networks that were reduced, also updates the reducible pairs and their heights.

        This method is similar to Reduce_Trivial_Pairs, but it uses and updates the dictionary of reducible pairs to reduce the trivial pairs.
        It also keeps track of the heights of the reduced pairs.

        :note: This method changes the set of networks in the problem, so it should only be called in a copy of the problem.
        """
        seq = []
        reduced_tree_sets = []
        heights_seq = []
        while candidate_leaves:
            l = candidate_leaves.pop()
            new_pairs = list(self.Trivial_Pair_With(l))
            if new_pairs:
                seq += new_pairs
                for p in new_pairs:
                    height_p = self.Height_Pair(p, reducible_pairs[p])
                    red_trees_p = self.Reduce_Pair_In_All(
                        p, reducible_pairs=reducible_pairs
                    )
                    heights_seq += [height_p]
                    reducible_pairs = self.Update_Reducible_Pairs(
                        reducible_pairs, red_trees_p
                    )
                    reduced_tree_sets += [red_trees_p]
                    candidate_leaves = candidate_leaves | set(p)
        return seq, reduced_tree_sets, reducible_pairs, heights_seq

    def Trivial_Pair_With(self, l):
        """
        Returns all trivial pairs involving the leaf l as first element of the pair.
        """
        pairs = set()
        # Go through all trees t with index i.
        for i, t in self.trees.items():
            # If the leaf occurs in t
            if l in t.leaves:
                # Compute reducible pairs of t with the leaf as first coordinate
                pairs_in_t = set(find_reducible_pairs_with_first(t, l))
                # If we did not have a set of candidate pairs yet, use pairs_in_t
                if not pairs:
                    pairs = pairs_in_t
                # Else, the candidate pairs must also be in t, so take intersection
                else:
                    pairs = pairs & set(pairs_in_t)
                # If we do not have any candidate pairs after checking a tree with l as leaf, we stop.
                if not pairs:
                    break
        return pairs

    def Improve_Sequence(self, CPS, reduced_trees, progress=False):
        """
        Improves a sequence 'CPS' for the input trees by removing elements and checking whether the new sequence still reduces all trees.
        Returns this improved sequence and the corresponding sets of reduced trees for each pair.
        """
        seq = deepcopy(CPS)
        i = 0
        while i < len(seq):
            redundant = True
            relevant_tree_indices = reduced_trees[i]
            new_relevant_pairs_for_trees = dict()
            for j in relevant_tree_indices:
                # Check if the shorter sequence reduces the trees, and if so, record which pairs reduced a cherry in which tree
                new_relevant_pairs_for_trees[j] = get_indices_of_reducing_pairs(
                    seq[:i] + seq[i + 1 :],
                    self.trees[j],
                )
                if not new_relevant_pairs_for_trees[j]:
                    redundant = False
                    break
            if redundant:
                # Remove the ith element from seq and reduced trees
                seq.pop(i)
                reduced_trees.pop(i)
                # Update reduced_trees for the relevant trees
                # First remove all j in relevant_tree_indices
                for tree_index_set in reduced_trees:
                    for j in relevant_tree_indices:
                        tree_index_set.discard(j)
                # Now add them back at the right places, according to "new_relevant_pairs_for_trees"
                for j in relevant_tree_indices:
                    for index in new_relevant_pairs_for_trees[j]:
                        reduced_trees[index].add(j)
                if progress:
                    print("New length is " + str(len(seq)))
                    print("Continue at position " + str(i))
            else:
                i += 1
        return seq, reduced_trees



"""
Classes for phylogenetic networks.

This module contains methods for checking whether a network is in a certain class.
Such as binary, tree-based, tree-child, orchard, or stack-free.

Originally written by Remie Janssen for the paper
"Comparing the topology of phylogenetic network generators"
by Remie Janssen and Pengy Liu.
"""


from .dinetwork import (
    is_binary,
    is_tree_based,
    is_tree_child,
    is_orchard,
    is_stack_free,
)
from .networkclass import *



"""
A module for rearranging phylogenetic networks using rearrangement moves.

This module contains the class RearrangementProblem which is used to represent a rearrangement problem.
That is, a problem of transforming one phylogenetic network into another using rearrangement moves.
A solution to a rearrangement problem is a sequence of moves, which is a list of phylox.rearrangement.move.Move.

There are different types of horizontal and vertical moves, enumerated in phylox.rearrangement.move.MoveType.
For an overview of types of moves, and some results about the existence of solutions to rearrangement problems,
see for example "Rearranging Phylogentic Networks" by Remie Janssen.
The code for this module was originally written by Remie Janssen for the above mentioned thesis.
"""



"""
A module containing the Green Line heuristic and its random version.
The Green Line heuristic is a heuristic to find a sequence of rSPR or tail moves between two phylogenetic networks.
The algorithms referred to in the comments are from the R Janssen's PhD thesis, "Rearranging Phylogenetic Networks", 2021.
"""

from phylox.rearrangement.heuristics.utils import *
from phylox.rearrangement.movability import check_valid, check_movable
from phylox.rearrangement.move import Move, apply_move, MoveType, apply_move_sequence
from phylox.exceptions import InvalidMoveException, InvalidMoveDefinitionException
from networkx.utils.decorators import py_random_state
from phylox.classes.dinetwork import is_binary, is_leaf_labeled_single_root_network


@py_random_state("seed")
def _GL_Case1_rSPR(N, Np, up, isom_N_Np, isom_Np_N, randomNodes=False, seed=None):
    """
    An implementation of Algorithm 2. Finds a sequence of rSPR moves that makes it possible to add the lowest reticulation up to the down-closed isomrophism.

    :param N: a phylogenetic network.
    :param Np: a phylogenetic network.
    :param up: a lowest reticulation node of Np above the isomorphism.
    :param isom_N_Np: a dictionary, containing a partial (down-closed) isomorphism map from N to Np. The inverse of isom_Np_N.
    :param isom_Np_N: a dictionary, containing a partial (down-closed) isomorphism map from Np to N. The inverse of isom_N_Np.
    :param randomNodes: a boolean value, determining whether the random version of this lemma is used.
    :param seed: a seed for the random number generator.
    :return: a list of rSPR moves in N, a list of rSPR moves in Np, a node of N, a node of Np. After performing the lists of moves on the networks, the nodes can be added to the isomorphism.
    """
    # use notation as in the paper
    # ' is denoted p
    xp = Np.child(up)
    x = isom_Np_N[xp]
    z = N.parent(x, exclude=isom_N_Np.keys(), randomNodes=randomNodes, seed=seed)

    # Case1a: z is a reticulation
    if N.in_degree(z) == 2:
        return [], [], z, up
    # Case1b: z is not a reticulation
    # Find a retic v in N not in the isom yet
    v = FindRetic(N, excludedSet=isom_N_Np.keys(), randomNodes=randomNodes, seed=seed)
    u = None
    for parent in N.predecessors(v):
        try:
            move = Move(
                move_type=MoveType.HEAD,
                moving_edge=(parent, v),
                target=(z, x),
                network=N,
            )
            check_valid(N, move)
        except (InvalidMoveException, InvalidMoveDefinitionException):
            continue
        if not randomNodes:
            u = parent
            break
        elif u == None or seed.getrandbits(1):
            u = parent
    # If v has an incoming arc (u,v) movable to (z,x), perform this move
    if u != None:
        move = Move(
            move_type=MoveType.HEAD, moving_edge=(u, v), target=(z, x), network=N
        )
        return [move], [], v, up
    # if none of the moves is valid, this must be because
    # v is already a reticulation above x with its movable incoming arc (u,v) with u=z.
    return [], [], v, up


@py_random_state("seed")
def GL_Case1_Tail(N, Np, up, isom_N_Np, isom_Np_N, randomNodes=False, seed=None):
    """
    An implementation of Algorithm 6. Finds a sequence of tail moves that makes it possible to add the lowest reticulation up to the down-closed isomrophism.

    :param N: a phylogenetic network.
    :param Np: a phylogenetic network.
    :param up: a lowest reticulation node of Np above the isomorphism.
    :param isom_N_Np: a dictionary, containing a partial (down-closed) isomorphism map from N to Np. The inverse of isom_Np_N.
    :param isom_Np_N: a dictionary, containing a partial (down-closed) isomorphism map from Np to N. The inverse of isom_N_Np.
    :param randomNodes: a boolean value, determining whether the random version of this lemma is used.
    :param seed: a seed for the random number generator.
    :return: a list of tail moves in N, a list of tail moves in Np, a node of N, a node of Np. After performing the lists of moves on the networks, the nodes can be added to the isomorphism. Returns false if the networks are not isomorphic with 2 leaves and 1 reticulation.
    """
    # use notation as in the paper
    # ' is denoted p
    xp = Np.child(up)
    x = isom_Np_N[xp]
    z = N.parent(x, exclude=isom_N_Np.keys(), randomNodes=randomNodes, seed=seed)
    # Case1a: z is a reticulation
    if N.in_degree(z) == 2:
        return [], [], z, up
    # Case1b: z is not a reticulation
    # z is a tree node
    if check_movable(N, (z, x), z):
        # Case1bi: (z,x) is movable
        # Find a reticulation u in N not in the isomorphism yet
        # TODO: Can first check if the other parent of x suffices here, should heuristcally be better
        u = FindRetic(N, excludedSet=isom_N_Np.keys(), randomNodes=randomNodes)
        v = N.child(u)
        if v == x:
            return [], [], u, up
        # we may now assume v!=x
        if z == v:
            v = N.child(z, exclude=[x])
            w = N.parent(u, randomNodes=randomNodes, seed=seed)
            move = Move(
                move_type=MoveType.TAIL,
                moving_edge=(z, v),
                target=(w, u),
                network=N,
            )
            return [move], [], u, up
        w = Parent(N, u, exclude=[z], randomNodes=randomNodes)
        move1 = Move(
            move_type=MoveType.TAIL, moving_edge=(z, x), target=(u, v), network=N
        )
        move2 = Move(
            move_type=MoveType.TAIL, moving_edge=(z, v), target=(w, u), origin=(u, x)
        )
        return [(move1, move2)], [], u, up
    # Case1bii: (z,x) is not movable
    c = N.parent(z)
    d = N.child(z, exclude=[x])
    # TODO: b does not have to exist if we have an outdeg-2 root, this could be c!
    b = N.parent(c)
    if N.in_degree(b) != 0:
        # Case1biiA: b is not the root of N
        a = N.parent(b, randomNodes=randomNodes, seed=seed)
        # First do the move ((c,d),c,(a,b)), then Case1bi applies as (z,x) is now movable
        move1 = Move(
            move_type=MoveType.TAIL, moving_edge=(c, d), target=(a, b), network=N
        )
        newN = apply_move(N, move1)
        u = FindRetic(
            newN, excludedSet=isom_N_Np.keys(), randomNodes=randomNodes, seed=seed
        )
        v = newN.child(u)
        if v == x:
            # In this case, u is a reticulation parent of x and u is not in the isom. Hence, we can simply add it to the isom.
            # Note: The tail move we did is not necessary!
            # TODO: First check both parents of x for being a reticulation not in the isomorphism yet
            return [], [], u, up
        # we may now assume v!=x
        if z == v:
            # This only happens if z==v and u==b
            # we move z up above the retic b as well, too
            w = newN.parent(b, randomNodes=randomNodes, seed=seed)
            move2 = Move(
                move_type=MoveType.TAIL, moving_edge=(z, d), target=(w, b), network=newN
            )
            return [move1, move2], [], u, up
        w = newN.parent(u, exclude=[z], randomNodes=randomNodes, seed=seed)
        move2 = Move(
            move_type=MoveType.TAIL, moving_edge=(z, x), target=(u, v), network=newN
        )
        move3 = Move(
            move_type=MoveType.TAIL, moving_edge=(z, v), target=(w, u), origin=(u, x)
        )
        return [move1, move2, move3], [], u, up
    # Case1biiB: b is the root of N
    # Note: d is not in the isomorphism
    e = N.child(d)
    if e == x:
        return [], [], d, up
    if N.out_degree(x) == 2:
        s = N.child(x)
        t = N.child(x, exclude=[s])
        if s == e:
            move = Move(
                move_type=MoveType.TAIL, moving_edge=(x, t), target=(d, e), network=N
            )
            return [move], [], d, up
        if t == e:
            move = Move(
                move_type=MoveType.TAIL, moving_edge=(x, s), target=(d, e), network=N
            )
            return [move], [], d, up
        moves = [
            Move(move_type=MoveType.TAIL, moving_edge=(x, s), target=(d, e), network=N),
            Move(
                move_type=MoveType.TAIL,
                moving_edge=(x, e),
                target=(z, t),
                origin=(d, s),
            ),
            Move(
                move_type=MoveType.TAIL,
                moving_edge=(x, t),
                target=(d, s),
                origin=(z, e),
            ),
        ]
        return moves, [], d, up
    if N.out_degree(e) == 2:
        s = N.child(e)
        t = N.child(e, exclude=[s])
        if s == x:
            move = Move(
                move_type=MoveType.TAIL, moving_edge=(e, t), target=(z, x), network=N
            )
            return [move], [], d, up
        if t == x:
            move = Move(
                move_type=MoveType.TAIL, moving_edge=(e, s), target=(z, x), network=N
            )
            return [move], [], d, up
        moves = [
            Move(move_type=MoveType.TAIL, moving_edge=(e, s), target=(z, x), network=N),
            Move(
                move_type=MoveType.TAIL,
                moving_edge=(e, x),
                target=(d, t),
                origin=(z, s),
            ),
            Move(
                move_type=MoveType.TAIL,
                moving_edge=(e, t),
                target=(z, s),
                origin=(d, x),
            ),
        ]
        return moves, [], d, up
    # neither are tree nodes, so both must be leaves
    # In that case, there is no sequence between the two networks.
    return [], [], None, None


@py_random_state("seed")
def GL_Case3(N, Np, up, isom_N_Np, isom_Np_N, randomNodes=False, seed=None):
    """
    An implementation of Algorithm 3. Finds a sequence of tail moves that makes it possible to add the lowest tree node up to the down-closed isomrophism.

    :param N: a phylogenetic network.
    :param Np: a phylogenetic network.
    :param up: a lowest tree node of Np above the isomorphism.
    :param isom_N_Np: a dictionary, containing a partial (down-closed) isomorphism map from N to Np. The inverse of isom_Np_N.
    :param isom_Np_N: a dictionary, containing a partial (down-closed) isomorphism map from Np to N. The inverse of isom_N_Np.
    :param randomNodes: a boolean value, determining whether the random version of this lemma is used.
    :param seed: a seed for the random number generator.
    :return: a list of tail moves in N, a list of tail moves in Np, a node of N, a node of Np. After performing the lists of moves on the networks, the nodes can be added to the isomorphism.
    """
    # Find the children x' and y' of u'
    xp, yp = list(Np.successors(up))
    # Find the corresponding nodes x and y in N
    x = isom_Np_N[xp]
    y = isom_Np_N[yp]
    # Find the set of common parents of x and y
    parents_x = set(N.predecessors(x))
    parents_y = set(N.predecessors(y))
    common_parents = parents_x & parents_y
    # Case3a: x and y have a common parent not in the isom
    common_parents_not_Y = []
    for parent in common_parents:
        if parent not in isom_N_Np.keys():
            # then parent can be mapped to up
            common_parents_not_Y += [parent]
            if not randomNodes:
                return [], [], parent, up
    if common_parents_not_Y:
        return [], [], seed.choice(common_parents_not_Y), up

    # Case3b: x and y do not have a common parent in the isomorphism
    # For both, find a parent not in the isomorphism yet
    # TODO: preferably make them tree nodes
    z_x = N.parent(x, exclude=isom_N_Np.keys(), randomNodes=randomNodes, seed=seed)
    z_y = N.parent(y, exclude=isom_N_Np.keys(), randomNodes=randomNodes, seed=seed)

    # Case3bi: (z_x,x) is movable
    try:
        move = Move(
            move_type=MoveType.TAIL, moving_edge=(z_x, x), target=(z_y, y), network=N
        )
        check_valid(N, move)
        return [move], [], z_x, up
    except (InvalidMoveException, InvalidMoveDefinitionException):
        pass
    # Case3bii: (z_y,y) is movable
    try:
        move = Move(
            move_type=MoveType.TAIL, moving_edge=(z_y, y), target=(z_x, x), network=N
        )
        check_valid(N, move)
        return [move], [], z_y, up
    except (InvalidMoveException, InvalidMoveDefinitionException):
        pass
    # Case3biii: Neither (z_x,x) nor (z_y,y) is movable

    if N.in_degree(z_x) == 2 or N.in_degree(z_y) == 2:
        return [], [], None, None
    # Both these parents are tree nodes.
    # This happens always in the non-random version, as otherwise there is a lowest reticulation node and we would be in Case 1 or 2.

    # As both nodes are tree nodes and the arcs immovable, both arcs hang of the side of a triangle.
    # Find the top node of the triangle for z_x
    c_x = N.parent(z_x)
    b_x = N.parent(c_x)

    # Find the top node of the triangle for z_y
    c_y = N.parent(z_y)
    b_y = N.parent(c_y)

    if N.in_degree(b_x) == 0:
        # c_x is the child of the root
        # c_x!=c_y, so c_y is not the child of the root
        # swap the roles of x and y
        x, y = y, x
        z_x, z_y = z_y, z_x
        b_x, b_y = b_y, b_x
        c_x, c_y = c_y, c_x
    # c_x is not the child of the root
    # find a parent of b_x, and the bottom node of the triangle d_x
    a_x = N.parent(b_x, randomNodes=randomNodes, seed=seed)
    d_x = N.child(c_x, exclude=[z_x])
    # e_x not in the proof, but needed to define the origin of the second move
    e_x = N.parent(c_x)
    moves = [
        Move(
            move_type=MoveType.TAIL,
            moving_edge=(c_x, d_x),
            target=(a_x, b_x),
            network=N,
        ),
        Move(
            move_type=MoveType.TAIL,
            moving_edge=(z_x, x),
            target=(z_y, y),
            origin=(e_x, d_x),
        ),
    ]
    return moves, [], z_x, up


class HeuristicDistanceMixin:
    """
    A class containing the Green Line heuristic and its random version.
    Meant to be inherited by the RearrangementProblem class.
    """

    def check_green_line_requirements(self):
        if not self.move_type in [MoveType.TAIL, MoveType.RSPR, MoveType.ALL]:
            raise Exception("Move type not supported by Green Line heuristic")
        if not is_binary(self.network1) or not is_binary(self.network2):
            raise Exception("Green Line heuristic only works for binary networks")
        if not is_leaf_labeled_single_root_network(
            self.network1
        ) or not is_leaf_labeled_single_root_network(self.network2):
            raise Exception(
                "Green Line heuristic only works for leaf-labeled networks with a single root"
            )
        if not len(self.network1.leaves) == len(self.network1.labels) or not len(
            self.network2.leaves
        ) == len(self.network2.labels):
            raise Exception(
                "Green Line heuristic only works for networks with unique labels"
            )
        if self.network1.labels == self.network2.labels:
            raise Exception(
                "Green Line heuristic only works for networks with the same set of labels"
            )

    def heuristic_green_line(self):
        """
        An implementation of Algorithm 4 and its tail move counterpart. Finds a sequence of tail/rSPR moves from network1 to network2 by building a down-closed isomorphism.
        Assumes the networks have the same leaf set, the same number of reticulations, are both binary, and all labels are unique.

        :return: A list of tail/rSPR moves from network1 to network2. Returns False if such a sequence does not exist.
        """
        self.check_green_line_requirements()

        head_moves = self.move_type in [MoveType.RSPR, MoveType.ALL]

        # Find the root and labels of the networks
        root1 = list(self.network1.roots)[0]
        root2 = list(self.network2.roots)[0]

        # initialize isomorphism
        isom_1_2 = dict()
        isom_2_1 = dict()
        isom_size = 0
        for label, [node1] in self.network1.labels.items():
            node2 = self.network2.labels[label][0]
            isom_1_2[node1] = node2
            isom_2_1[node2] = node1
            isom_size += 1

        # Keep track of the size of the isomorphism and the size it is at the end of the green line algorithm
        goal_size = len(self.network1) - 1

        # init lists of sequence of moves
        # list of (moving_edge,moving_endpoint,from_edge,to_edge)
        seq_from_1 = []
        seq_from_2 = []
        # TODO keep track of lowest nodes? (Currently, the code does not do this, but it could speed up the code)

        network1 = self.network1
        network2 = self.network2

        # Do the green line algorithm
        while isom_size < goal_size:
            # Find lowest nodes above the isom in the networks:
            (
                lowest_tree_node_network1,
                lowest_retic_network1,
            ) = LowestReticAndTreeNodeAbove(network1, isom_1_2.keys())
            (
                lowest_tree_node_network2,
                lowest_retic_network2,
            ) = LowestReticAndTreeNodeAbove(network2, isom_2_1.keys())

            ######################################
            # Case1: a lowest retic in network1
            if lowest_retic_network1 != None:
                # use notation as in the paper network1 = N', network2 = N, where ' is denoted p
                up = lowest_retic_network1
                if head_moves:
                    (
                        moves_network_2,
                        moves_network_1,
                        added_node_network_2,
                        added_node_network_1,
                    ) = _GL_Case1_rSPR(network2, network1, up, isom_2_1, isom_1_2)
                else:
                    (
                        moves_network_2,
                        moves_network_1,
                        added_node_network_2,
                        added_node_network_1,
                    ) = GL_Case1_Tail(network2, network1, up, isom_2_1, isom_1_2)
                    if added_node_network_1 == None:
                        return False
            ######################################
            # Case2: a lowest retic in network2
            elif lowest_retic_network2 != None:
                # use notation as in the paper network2 = N', network1 = N, where ' is denoted p
                up = lowest_retic_network2
                if head_moves:
                    (
                        moves_network_1,
                        moves_network_2,
                        added_node_network_1,
                        added_node_network_2,
                    ) = _GL_Case1_rSPR(network1, network2, up, isom_1_2, isom_2_1)
                else:
                    (
                        moves_network_1,
                        moves_network_2,
                        added_node_network_1,
                        added_node_network_2,
                    ) = GL_Case1_Tail(network1, network2, up, isom_1_2, isom_2_1)
                    if added_node_network_1 == None:
                        return False

                        ######################################
            # Case3: a lowest tree node in network1
            else:
                # use notation as in the paper network1 = N, network2 = N'
                up = lowest_tree_node_network2
                (
                    moves_network_1,
                    moves_network_2,
                    added_node_network_1,
                    added_node_network_2,
                ) = GL_Case3(network1, network2, up, isom_1_2, isom_2_1)
            # Now perform the moves and update the isomorphism
            isom_1_2[added_node_network_1] = added_node_network_2
            isom_2_1[added_node_network_2] = added_node_network_1
            seq_from_1 += moves_network_1
            seq_from_2 += moves_network_2
            network1 = apply_move_sequence(network1, moves_network_1)
            network2 = apply_move_sequence(network2, moves_network_2)
            isom_size += 1

        # Add the root to the isomorphism, if it was there
        isom_1_2[root1] = root2
        isom_2_1[root2] = root1
        # invert seq_from_2, rename to node names of network1, and append to seq_from_1
        return seq_from_1 + [
            move.invert().rename_nodes(isom_2_1) for move in reversed(seq_from_2)
        ]

    @py_random_state("seed")
    def heuristic_green_line_random(self, seed=None):
        """
        An implementation of Algorithm 5 and its tail move counterpart. Finds a sequence of tail/rSPR moves from network1 to network2 by randomly building a down-closed isomorphism.

        :return: A list of tail/rSPR moves from network1 to network2. Returns False if such a sequence does not exist.
        """
        self.check_green_line_requirements()
        head_moves = self.move_type in [MoveType.RSPR, MoveType.ALL]

        # Find the root and labels of the networks
        root1 = list(self.network1.roots)[0]
        root2 = list(self.network2.roots)[0]

        # initialize isomorphism
        isom_1_2 = dict()
        isom_2_1 = dict()
        isom_size = 0
        for label, [node1] in self.network1.labels.items():
            node2 = self.network2.labels[label][0]
            isom_1_2[node1] = node2
            isom_2_1[node2] = node1
            isom_size += 1

        # Keep track of the size of the isomorphism and the size it is at the end of the green line algorithm
        goal_size = len(self.network1) - 1

        # init lists of sequence of moves
        # list of (moving_edge,moving_endpoint,from_edge,to_edge)
        seq_from_1 = []
        seq_from_2 = []
        # TODO keep track of lowest nodes? (Currently, the code does not do this, but it could speed up the code)

        network1 = self.network1
        network2 = self.network2

        # Do the green line algorithm
        while isom_size < goal_size:
            # Find all lowest nodes above the isom in the networks:
            (
                lowest_tree_node_network1,
                lowest_retic_network1,
            ) = LowestReticAndTreeNodeAbove(network1, isom_1_2.keys(), allnodes=True)
            (
                lowest_tree_node_network2,
                lowest_retic_network2,
            ) = LowestReticAndTreeNodeAbove(network2, isom_2_1.keys(), allnodes=True)

            # Construct a list of all lowest nodes in a tuple with the corresponding network (in random order)
            # I.e. If u is a lowest node of network one, it will appear in the list as (u,1)
            lowest_nodes_network1 = [
                (u, 1) for u in lowest_tree_node_network1 + lowest_retic_network1
            ]
            lowest_nodes_network2 = [
                (u, 2) for u in lowest_tree_node_network2 + lowest_retic_network2
            ]
            candidate_lowest_nodes = lowest_nodes_network1 + lowest_nodes_network2
            seed.shuffle(candidate_lowest_nodes)

            # As some cases do not give an addition to the isom, we continue trying lowest nodes until we find one that does.
            for lowest_node, network_number in candidate_lowest_nodes:
                ######################################
                # Case1: a lowest retic in network1
                if network_number == 1 and network1.in_degree(lowest_node) == 2:
                    # use notation as in the paper network1 = N', network2 = N, where ' is denoted p
                    up = lowest_node
                    if head_moves:
                        (
                            moves_network_2,
                            moves_network_1,
                            added_node_network_2,
                            added_node_network_1,
                        ) = _GL_Case1_rSPR(
                            network2,
                            network1,
                            up,
                            isom_2_1,
                            isom_1_2,
                            randomNodes=True,
                            seed=seed,
                        )
                    else:
                        (
                            moves_network_2,
                            moves_network_1,
                            added_node_network_2,
                            added_node_network_1,
                        ) = GL_Case1_Tail(
                            network2,
                            network1,
                            up,
                            isom_2_1,
                            isom_1_2,
                            randomNodes=True,
                            seed=seed,
                        )
                        if added_node_network_1 == None:
                            # The networks are non-isom networks with 2 leaves and 1 reticulation
                            return False
                    # This case always gives a node to add to the isom
                    break

                ######################################
                # Case2: a lowest retic in network2
                elif network_number == 2 and network2.in_degree(lowest_node) == 2:
                    # use notation as in the paper network2 = N', network1 = N, where ' is denoted p
                    up = lowest_node
                    if head_moves:
                        (
                            moves_network_1,
                            moves_network_2,
                            added_node_network_1,
                            added_node_network_2,
                        ) = _GL_Case1_rSPR(
                            network1,
                            network2,
                            up,
                            isom_1_2,
                            isom_2_1,
                            randomNodes=True,
                            seed=seed,
                        )
                    else:
                        (
                            moves_network_1,
                            moves_network_2,
                            added_node_network_1,
                            added_node_network_2,
                        ) = GL_Case1_Tail(
                            network1,
                            network2,
                            up,
                            isom_1_2,
                            isom_2_1,
                            randomNodes=True,
                            seed=seed,
                        )
                        if added_node_network_1 == None:
                            # The networks are non-isom networks with 2 leaves and 1 reticulation
                            return False
                            # This case always gives a node to add to the isom
                    break

                ######################################
                # Case3: a lowest tree node in network1
                elif network_number == 2 and network2.out_degree(lowest_node) == 2:
                    # use notation as in the paper network1 = N, network2 = N'
                    up = lowest_node
                    (
                        moves_network_1,
                        moves_network_2,
                        added_node_network_1,
                        added_node_network_2,
                    ) = GL_Case3(
                        network1,
                        network2,
                        up,
                        isom_1_2,
                        isom_2_1,
                        randomNodes=True,
                        seed=seed,
                    )
                    # If we can add a node to the isom, added_node_network_2 has a value
                    if added_node_network_2:
                        break

                ######################################
                # Case3': a lowest tree node in network2
                else:
                    # use notation as in the paper network1 = N, network2 = N'
                    up = lowest_node
                    (
                        moves_network_2,
                        moves_network_1,
                        added_node_network_2,
                        added_node_network_1,
                    ) = GL_Case3(
                        network2,
                        network1,
                        up,
                        isom_2_1,
                        isom_1_2,
                        randomNodes=True,
                        seed=seed,
                    )
                    # If we can add a node to the isom, added_node_network_2 has a value
                    if added_node_network_2:
                        break

            # Now perform the moves and update the isomorphism
            isom_1_2[added_node_network_1] = added_node_network_2
            isom_2_1[added_node_network_2] = added_node_network_1
            seq_from_1 += moves_network_1
            seq_from_2 += moves_network_2
            network1 = apply_move_sequence(network1, moves_network_1)
            network2 = apply_move_sequence(network2, moves_network_2)
            isom_size += 1

        # Add the root to the isomorphism, if it was there
        isom_1_2[root1] = root2
        isom_2_1[root2] = root1

        # invert seq_from_2, rename to node names of network1, and append to seq_from_1
        return seq_from_1 + [
            move.invert().rename_nodes(isom_2_1) for move in reversed(seq_from_2)
        ]



import time

from phylox.exceptions import NoSolutionException, TimeoutException
from phylox.isomorphism import is_isomorphic
from phylox.rearrangement.move import (
    Move,
    MoveType,
    all_valid_moves,
    apply_move_sequence,
)


class ExactMethodsMixin(object):
    def solve_depth_first(self, max_time=False, show_bounds=True):
        """
        An implementation of Algorithm 1 from R. Janssen's PhD thesis.
        Uses an iterated Depth First Search to simulate a Breath First Search.

        :param max_time: a float, a time limit for the function in seconds. If False, no time limit is used, and the function continues until it finds a sequence.
        :param show_bounds: a boolean parameter, if True the current lower bounds are printed to the terminal, used for debugging.
        :return: a shortest sequence of moves between the networks if it is found within the time limit, otherwise it returns an integer: a lower bound for the length of the shortest sequence between the networks.

        :example:
        >>> from phylox.rearrangement.rearrangementproblem import RearrangementProblem

        """
        lower_bound = 0
        stop_time = False
        if max_time:
            stop_time = time.time() + max_time
        while True:
            output = self.solve_depth_first_bounded(
                max_depth=lower_bound, stop_time=stop_time
            )
            if type(output) == list:
                break
            lower_bound += 1
            if show_bounds:
                print(lower_bound)
        return output

    # Finds a shortest sequence between network1 and network2 using DFS with bounded depth
    def solve_depth_first_bounded(self, max_depth=0, stop_time=False):
        """
        A subroutine of Algorithm 1 of R. Janssen's PhD thesis.
        A depth-bounded Depth First Search used to simulate
        a Breath First Search.

        :param max_depth: a integer, the maximum depth for the search tree.
        :param stop_time: a float, a time limit for the function in clock time.
            If False, no time limit is used, and the function continues until
            it finds a sequence.
        :return: a shortest sequence of at most max_depth moves between the
            networks if it is found before the stop_time, otherwise it returns an False.
        """
        # Make a stack and search
        stack = [[]]
        while stack:
            current_moves = stack.pop()
            current_length = len(current_moves)
            current_network = apply_move_sequence(self.network1, current_moves)
            if current_length == max_depth and is_isomorphic(
                current_network, self.network2
            ):
                return current_moves
            if current_length < max_depth:
                stack += [
                    current_moves + [move]
                    for move in all_valid_moves(current_network, self.move_type)
                ]
            if stop_time and time.time() > stop_time:
                raise TimeoutException
        return False


# # Finds a shortest sequence between network1 and network2 using BFS
# def Breadth_First(network1, network2, tail_moves=True, head_moves=True, max_time=False):
#     """
#     A true BFS implementation to find a shortest sequence between two networks. This implementation uses too much memory, use Depth_First!

#     :param network1: a phylogenetic network.
#     :param network2: a phylogenetic network.
#     :param tail_moves: a boolean value determining whether tail moves are used.
#     :param head_moves: a boolean value determining whether head moves are used.
#     :param max_time: a float, a time limit for the function in seconds. If False, no time limit is used, and the function continues until it finds a sequence.
#     :return: a shortest sequence of moves between the networks if it is found within the time limit, otherwise it returns an integer: a lower bound for the length of teh shortest sequence between the networks.
#     """
#     # If we cannot do any moves:
#     if not tail_moves and not head_moves:
#         if Isomorphic(network1, network2):
#             return 0
#         else:
#             return False
#     # Else, make a queue and search
#     queue = deque([[]])
#     start_time = time.time()
#     while queue:
#         current_moves = queue.popleft()
#         current_network = network1
#         for move in current_moves:
#             current_network = DoMove(current_network, *move)
#         if Isomorphic(current_network, network2):
#             return current_moves
#         validMoves = AllValidMoves(current_network, tail_moves=tail_moves, head_moves=head_moves)
#         for move in validMoves:
#             queue.append(current_moves + [move])
#         if max_time and time.time() - start_time > max_time:
#             return len(current_moves)
#     return False



"""
Adapted from script provided by (a colleage of) Pengyu Liu
which contains the code for Beta-splitting model (Aldous1996)

The beta-splitting model is a model for generating random binary trees.
The model is parameterized by a parameter beta > 0 which determines the shape of the tree.
"""

import numpy as np
from networkx.utils.decorators import np_random_state, py_random_state
from scipy.special import loggamma

from phylox import DiNetwork
from phylox.constants import LABEL_ATTR

############################################
# Simulation functions
############################################

# _a_n is a normalizing constant defined in
# Equation (2) of Aldous1996 (so the sum of
# the values is equal to 1. It is not
# computed here to save time.
def _a_n(beta):
    return 1


# Compute the "probability" to split n in (i|n-1), where i=1,..,n-1
def _compute_split_probability(n, beta):
    q_n = []
    for i in range(1, n):
        q_i_n = np.exp(
            (loggamma(beta + i + 1) + loggamma(beta + n - i + 1))
            - ((_a_n(beta) + loggamma(i + 1) + loggamma(n - i + 1)))
        )
        q_n.append(q_i_n)
    return q_n


@py_random_state("seed")
def simulate_beta_splitting(n, beta, seed=None):
    """
    Simulate a random binary tree with n leaves using the beta-splitting model.

    :param n: the number of leaves of the tree.
    :param beta: the beta parameter of the beta-splitting model.
    :return: a random binary tree with n leaves using the beta-splitting model.

    :example:
    >>> from phylox.generators.trees.beta_splitting_tree import simulate_beta_splitting
    >>> tree = simulate_beta_splitting(5, 1)
    >>> len(tree.leaves)
    5
    """
    # Initialize tree.
    tree = DiNetwork()
    tree.add_edge(-1, n + 1)
    tree.nodes[n + 1][LABEL_ATTR] = n
    last_internal_node = n + 1
    last_leaf_node = 0
    queue = [n + 1]
    # Insert one node at each iteration.
    while queue:
        node = queue.pop()
        n_node = tree.nodes[node].get(LABEL_ATTR)
        # Internal node. Splits again.
        if n_node > 1:
            # Compute the "probability" to split n in (i|n-1), where i=1,..,n-1
            q_n = _compute_split_probability(n_node, beta)
            split = seed.choices(population=list(range(1, n_node)), weights=q_n, k=1)[0]
            # Create children.
            for new_n in [split, n_node - split]:
                if new_n == 1:
                    tree.add_edge(node, last_leaf_node + 1)
                    last_leaf_node += 1
                else:
                    tree.add_edge(node, last_internal_node + 1)
                    tree.nodes[last_internal_node + 1][LABEL_ATTR] = new_n
                    queue.append(last_internal_node + 1)
                    last_internal_node += 1
    # Return tree.
    return tree



from networkx.utils.decorators import np_random_state, py_random_state

from phylox import DiNetwork

### For the tree child network containment paper by Murakami and Janssen
### This file contains functions to generate random tree-child (sub)networks


@py_random_state("seed")
def random_tree_child_sequence(leaves, reticulations, seed=None):
    """
    Returns a random tree-child sequence with a given number of leaves and reticulations
    :param leaves: number of leaves
    :param reticulations: number of reticulations
    :param seed: seed for the random number generator
    :return: a random tree-child sequence

    :example:
    >>> from phylox.generators.randomTC.random_tc_network import random_tree_child_sequence
    >>> seq = random_tree_child_sequence(5, 2)
    >>> len(seq)
    6
    """
    current_leaves = set([1, 2])
    seq = [(2, 1)]
    not_forbidden = set([2])
    leaves_left = leaves - 2
    retics_left = reticulations

    # Continue until we added enough leaves and reticulations
    while leaves_left > 0 or retics_left > 0:
        # Decide if we add a leaf, or a reticulation
        type_added = "L"
        if len(not_forbidden) > 0 and leaves_left > 0 and retics_left > 0:
            if (
                seed.randint(0, leaves_left + retics_left - 1) < retics_left
            ):  # probability of retic depends on number of reticulations left to add
                # if random.randint(0 , 1)<1:   #probability of reticulations and leaves are the same if both are an option
                type_added = "R"
        elif len(not_forbidden) > 0 and retics_left > 0:
            type_added = "R"
        elif leaves_left > 0:
            type_added = "L"
        else:
            return False

        # Actually add the pair
        if type_added == "R":
            first_element = seed.choice(list(not_forbidden))
            retics_left -= 1
        if type_added == "L":
            first_element = len(current_leaves) + 1
            leaves_left -= 1
            current_leaves.add(first_element)
            not_forbidden.add(first_element)

        second_element = seed.choice(list(current_leaves - set([first_element])))
        not_forbidden.discard(second_element)
        seq.append((first_element, second_element))

    # reverse the sequence, as it was built in reverse order
    seq = [pair for pair in reversed(seq)]
    return seq


@py_random_state("seed")
def random_tree_child_subsequence(seq, r, seed=None):
    """
    Returns a random tree-child subsequence with a given number of reticulations
    :param seq: a tree-child sequence
    :param r: number of reticulations in the subsequence
    :param seed: seed for the random number generator
    :return: a random tree-child subsequence

    :example:
    >>> from phylox.generators.randomTC.random_tc_network import random_tree_child_subsequence
    >>> seq = [(4,1), (4, 1), (3, 1), (2, 1), (2, 1), (2,1)]
    >>> newSeq = random_tree_child_subsequence(seq, 1)
    >>> len(newSeq)
    4
    """
    # First `uniformly at random' choose one pair per leaf, with that leaf as first element
    leaves = dict()
    indices = set()
    for i, pair in enumerate(seq):
        x = pair[0]
        if x not in leaves:
            indices.add(i)
            leaves[x] = (1, i)
        else:
            if seed.randint(0, leaves[x][0]) < 1:
                indices.remove(leaves[x][1])
                indices.add(i)
                leaves[x] = (leaves[x][0] + 1, i)
            else:
                leaves[x] = (leaves[x][0] + 1, leaves[x][1])
    # Add r reticulations with a max of the whole sequence
    unused = set(range(len(seq))) - indices
    for j in range(r):
        new = seed.choice(list(unused))
        unused = unused - set([new])
        indices.add(new)
    newSeq = []
    for i, pair in enumerate(seq):
        if i in indices:
            newSeq.append(pair)
    return newSeq


@py_random_state("seed")
def generate_network_random_tree_child_sequence(
    leaves, reticulations, label_leaves=False, seed=None
):
    """
    Returns a random tree-child network with a given number of leaves and reticulations
    :param leaves: number of leaves
    :param reticulations: number of reticulations
    :param label_leaves: whether to label the leaves
    :return: a random tree-child network

    :example:
    >>> from phylox.generators.randomTC.random_tc_network import generate_network_random_tree_child_sequence
    >>> network = generate_network_random_tree_child_sequence(5, 2)
    >>> len(network.leaves)
    5
    >>> network.reticulation_number
    2
    """
    if leaves < 2 or reticulations < 0:
        raise ValueError(
            "Invalid number of leaves or reticulations, must be at least 2 and 0 respectively"
        )
    seq = random_tree_child_sequence(leaves, reticulations, seed=seed)
    if not seq:
        return False
    return DiNetwork.from_cherry_picking_sequence(seq, label_leaves=label_leaves)



import os
import sys

import networkx as nx
import numpy as np
from networkx.utils.decorators import np_random_state

from phylox.base import find_unused_node
from phylox.exceptions import InvalidMoveDefinitionException, InvalidMoveException
from phylox.isomorphism import count_automorphisms
from phylox.rearrangement.invertsequence import from_edge
from phylox.rearrangement.move import Move, apply_move
from phylox.rearrangement.movetype import MoveType


def acceptance_probability(
    network,
    result_network,
    move,
    move_type_probabilities,
    number_of_leaves=None,
    current_reticulation_number=None,
    symmetries=False,
):
    """
    Computes the acceptance probability of a move.

    :param network: the network before the move.
    :param result_network: the network after the move.
    :param move: the move.
    :param move_type_probabilities: the move type probabilities.
    :param number_of_leaves: the number of leaves in the network.
    :param current_reticulation_number: the current number of reticulations in the network.
    :param symmetries: whether to correct for symmetries.
    :return: the acceptance probability of the move.
    """
    current_reticulation_number = (
        current_reticulation_number or network.reticulation_number
    )
    number_of_leaves = number_of_leaves or len(network.leaves)
    p = 0
    if move.move_type in [MoveType.TAIL, MoveType.HEAD]:
        p = 1
    if move.move_type == MoveType.VPLU:
        no_edges_network = float(
            2 * number_of_leaves + 3 * current_reticulation_number - 1
        )
        no_edges_network_after = no_edges_network + 3
        p = (
            (
                move_type_probabilities[MoveType.VMIN]
                / move_type_probabilities[MoveType.VPLU]
            )
            * no_edges_network**2
            / (no_edges_network_after)
        )
    if move.move_type == MoveType.VMIN:
        no_edges_network = float(
            2 * number_of_leaves + 3 * current_reticulation_number - 1
        )
        no_edges_network_after = no_edges_network - 3
        if no_edges_network > 3:
            p = (
                (
                    move_type_probabilities[MoveType.VPLU]
                    / move_type_probabilities[MoveType.VMIN]
                )
                * no_edges_network
                / (no_edges_network_after**2)
            )
    if symmetries:
        # correct for number of representations, i.e., symmetries.
        p *= count_automorphisms(network) / count_automorphisms(result_network)
    return p


@np_random_state("seed")
def sample_mcmc_networks(
    starting_network,
    move_type_probabilities,
    restriction_map=None,
    correct_symmetries=True,
    burn_in=1000,
    number_of_samples=1,
    add_root_if_necessary=False,
    seed=None,
):
    """
    Samples phylogenetic networks using a Markov-Chain Monte Carlo method.

    :param starting_network: the phylox.DiNetwork used as the starting point of the Markov chain.
    :param move_type_probabilities: a dictionary mapping MoveTypes to probabilities.
    :param restriction_map: a boolean function that takes a phylox.DiNetwork as input.
    :param correct_symmetries: whether to correct for symmetries in the acceptance probability, set to True for uniform distribution.
    :param burn_in: the number of steps (including rejected proposals) between each sample.
    :param number_of_samples: the number of networks to sample.
    :param add_root_if_necessary: whether to add a root edge to each root if it has out-degree > 1.
    :param seed: the seed for the random number generator.

    :return: a list of phylox.DiNetwork objects.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.movetype import MoveType
    >>> from phylox.generators.mcmc import sample_mcmc_networks
    >>> starting_network = DiNetwork(
    ...     edges = ((0,1), (0,2), (1,2), (1,3), (2,4)),
    ...     labels = ((3, "A"), (4, "B")),
    ... )
    >>> move_type_probabilities = {
    ...     MoveType.TAIL: 0.2,
    ...     MoveType.VPLU: 0.4,
    ...     MoveType.VMIN: 0.4,
    ... }
    >>> restriction_map = (lambda nw: nw.reticulation_number < 2)
    >>> sampled_networks = sample_mcmc_networks(
    ...     starting_network,
    ...     move_type_probabilities,
    ...     restriction_map=restriction_map,
    ...     correct_symmetries=False,
    ...     burn_in=100,
    ...     number_of_samples=50,
    ...     add_root_if_necessary=False,
    ...     seed=1,
    ... )
    >>> all([network.reticulation_number<2 for network in sampled_networks])
    True
    >>> all([len(network.leaves)==2 for network in sampled_networks])
    True
    """
    network = starting_network.copy()
    current_reticulation_number = network.reticulation_number
    number_of_leaves = len(network.leaves)
    if add_root_if_necessary:
        for root in network.roots:
            if network.out_degree(root) >= 1:
                new_root = find_unused_node(network)
                network.add_edges_from([(new_root, root)])
                root = new_root
        roots = network._set_roots()
    available_reticulations = set()
    available_tree_nodes = set()

    sample = []

    for _ in range(number_of_samples):
        non_moves = 0
        for _ in range(burn_in):
            try:
                move = Move.random_move(
                    network,
                    available_tree_nodes=available_tree_nodes,
                    available_reticulations=available_reticulations,
                    move_type_probabilities=move_type_probabilities,
                    seed=seed,
                )
                result_network = apply_move(network, move)
            except (InvalidMoveException, InvalidMoveDefinitionException) as e:
                non_moves += 1
                continue
            if seed.random() > acceptance_probability(
                network,
                result_network,
                move,
                move_type_probabilities,
                number_of_leaves=number_of_leaves,
                current_reticulation_number=current_reticulation_number,
                symmetries=correct_symmetries,
            ):
                non_moves += 1
                continue
            # only apply the move if the restrinction_map returns True
            if not (restriction_map is None or restriction_map(result_network)):
                non_moves += 1
                continue
            if move.move_type in [MoveType.TAIL, MoveType.HEAD]:
                network = result_network
            if move.move_type == MoveType.VPLU:
                current_reticulation_number += 1
                available_tree_nodes.discard(move.start_node)
                available_reticulations.discard(move.end_node)
                network = result_network
            if move.move_type == MoveType.VMIN:
                current_reticulation_number -= 1
                available_tree_nodes.add(move.removed_edge[0])
                available_reticulations.add(move.removed_edge[1])
                network = result_network
        sample.append(network)
    return sample
